#!/bin/bash

#   FILE: backup.sh -- Backup system to a CD-RW.
# AUTHOR: W. Michael Petullo <burn_baby_burn@flyn.org>
#   DATE: 20 SEP 1998
#
# Copyright (c) 1999 W. Michael Petullo <burn_baby_burn@flyn.org>
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

. /etc/burn_baby_burn/bbb_functions

TYPE=$1
ERR_DEST=$2
BOOTABLE=$3
VOL_ID="backup -- `date +'%d %b %Y'`"
USAGE="
Usage: ${0##*/} existing_media|new_media|cdr|<image path> log|stderr [bootable]\n
\n
Allows one to perform a backup to CD-R[W] media.  The directories to burn
are specified in /etc/burn_baby_burn/burn_baby_burn.  This program can
make bootable (El Torito) backup media.  If you want to boot a kernel
off of a CD, make sure your CD-ROM driver and the ISO-9660 filesystem
are not kernel modules.  Place a boot disk image in /BOOT_IMG or try
running mkboot.sh to make the necessary boot disk image. \n
\n
An inittab and rc file are provided which try to provide a sane environment,
given the root disk is read-only.  Backup.sh insures these are placed
in appropriate places in the generated ISO-9660 filesystem. \n
\n
existing_media checks for the existance of the file backup in the root
directory of the media being burned to.  This can allow one to use
${0##*/} in a cron script, without having to worry about accidentally
destroying the wrong disk.\n
new_media makes no such check.\n
cdr does not make a check or attempt to blank the disc. \n
Anything else causes an image file to be written to the path given. \n
\n
log causes error messages to be logged using the logger command. \n
stderr causes error messages to be printed to stderr. \n
\n
bootable causes backup.sh to attempt to create a bootable CD-ROM.\n"

trap 'rm -rf $TMP_DIR; rm -f /tmp/be_nice; exit 6' SIGHUP SIGINT SIGTERM
trap 'rm -rf $TMP_DIR; rm -f /tmp/be_nice' EXIT

if [ $# -lt 2 ]; then
	usage "$USAGE"
elif [ $ERR_DEST != log -a $ERR_DEST != stderr ]; then
	usage "$USAGE"
elif ! configured; then
	print_message $ERR_DEST "No config found in /etc or ~"
	exit 1
fi

if [ $TYPE = cdr ]; then
	MEDIA_TYPE=cdr
elif [ $TYPE = existing_media -o $TYPE = new_media ]; then
	MEDIA_TYPE=cdrw
else
	MEDIA_TYPE=$TYPE
fi

# Okay, there are two problems with this /tmp/be_nice system.  
#
# One, this lockfile does not affect CPU intensive cron jobs that begin
# running before backup.sh does.  Because of this, it is important to
# structure your crontab file accordingly.  For example, the scripts
# in /etc/cron.daily seem to run every day at 0400.  It makes sense to
# run backup.sh at 0355, and wrap all of the scripts in /etc/cron.daily
# in do_nicely.sh.

touch /tmp/be_nice
touch /tmp/backup_failed
print_message $ERR_DEST "Backup start time: `date`"

# Because this is meant to be run from a cron script, we need to make sure we 
# won't be overwriting the wrong stuff (ie: an mp3 CD).  The solution I am 
# currently using is to check for the existance of a file called 
# ``backup'' at the root of the filesystem.  Specifying new_media on the 
# command line skips this test.
	
if [ $TYPE = existing_media ]; then
	if ! disc_mounts; then
		print_message $ERR_DEST "WARNING: bad filesystem -- before burn"
		exit 2
	elif ! is_backup_disc; then
		print_message $ERR_DEST "WARNING: incorrect disc"
		exit 4
	fi
else
	umount $CD_DEV 2> /dev/null	
fi

# Alrighty, everything should okay, lets give it a (LASER) shot -- heh heh.

# Although this copy is slow and seems redundant, it may be necessary
# to ensure data does not change from time of burn to time of verify.
TMP_DIR=`mktemp -u /tmp/backup_tmp_dir.XXXXXX` \
	|| ( 
		print_message $ERR_DEST "WARNING: can\'t name tmp dir"
		exit 5 
	)
mkdir $TMP_DIR \
	|| ( 
		print_message $ERR_DEST "WARNING: can\'t mkdir tmp dir";
		exit 5
	)

# Build exclude arguments for rsync (SKIP_ARGS and SKIP_COMPRESS_ARGS) and 
# tar (SKIP_ARGS).
SKIP_ARGS=""
for dir in $SKIP_DIRS; do
	SKIP_ARGS="$SKIP_ARGS --exclude=\"$dir\""
done
for dir in `ls $HOME_DIR`; do
	SKIP_ARGS="$SKIP_ARGS --exclude=\"$HOME_DIR/$dir/no_backup\""
done
SKIP_COMRESS_ARGS=""
for dir in $COMPRESS_DIRS; do
	SKIP_COMPRESS_ARGS="$SKIP_COMPRESS_ARGS --exclude=\"$dir\""
done

# Copy data to temp dir.
eval rsync -a $SKIP_ARGS $SKIP_COMPRESS_ARGS $BACKUP_DIRS $TMP_DIR
for dir in $COMPRESS_DIRS; do
	CAR=`basename $dir`	# The path's car, when parsing from the right.
	CDR=`dirname $dir`	# The path's cdr, when parsing from the right.
	mkdir -p $TMP_DIR/$CDR
	( 
		cd $CDR; 
		tar cf - $CAR $SKIP_ARGS | bzip2 -c > $TMP_DIR/$CDR/$CAR.tar.bz2
	)
done
touch $TMP_DIR/backup

if ! burn_dir.sh "$VOL_ID" $TMP_DIR $MEDIA_TYPE log $BOOTABLE; then
	print_message $ERR_DEST "WARNING: error burning to media"
	exit 8
fi

# Backup seems to be okay.
rm -f /tmp/backup_failed
print_message $ERR_DEST "Backup end time: `date` size: $FS_SIZE"
exit 0
