#!/bin/bash

#   FILE: burn_dir.sh -- Burns the files in a directory.
# AUTHOR: W. Michael Petullo <burn_baby_burn@flyn.org>
#   DATE: 12 SEP 1998
#
# Copyright (c) 1999 W. Michael Petullo <burn_baby_burn@flyn.org>
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

. /etc/burn_baby_burn/bbb_functions

SRC_DIR=$2
MEDIA_TYPE=$3
ERR_DEST=$4

USAGE="Usage: ${0##*/} volume_name dir cdr|cdrw|<image path> log|stderr [bootable] \n
\n
Burns the files in a directory to CD-R[W] media as an ISO-9660
filesystem.\n
\n
cdr causes the media to be burned at a faster speed. \n
cdrw causes the media to be erased and burned at a slower speed. \n
Anything else causes an image file to be written to the path given. \n
\n
log causes error messages to be logged using the logger command. \n
stderr causes error messages to be printed to stderr.\n
\n
If CDRECORD_DEV is set to the empty string, burn_dir.sh will simply
create an image file and place it in /tmp.  This is useful for machines
with no CD-R drive."

if [ $# -lt 4 ]; then
	usage "$USAGE"
elif ! configured; then
	print_message $ERR_DEST "No config found in /etc or ~"
	exit 1
fi

if [ $MEDIA_TYPE = cdr ]; then
	SPEED=$CDR_SPEED
	BURN_IT="true"
elif [ $MEDIA_TYPE = cdrw ]; then
	SPEED=$CDRW_SPEED
	BURN_IT="true"
else
	IMAGE_FILE=$MEDIA_TYPE
	unset BURN_IT
fi

if [ "$5" = bootable ]; then
	BOOTABLE=true
fi

# This is a lock file of sorts.  Other CPU intensive scripts should check 
# for its existence before running.  Starving cdrecord of CPU cycles can 
# result in a corrupted CD.
touch /tmp/be_nice
# saf_t_kill_xscreensaver.sh

if too_much_data $SRC_DIR; then
	print_message $ERR_DEST "WARNING: filesystem too large"
	rm -f /tmp/be_nice
	exit 7
fi

if [ $# -gt 4 -a "$BOOTABLE" ]; then
	setup_boot_dirs $SRC_DIR
	if [ -f $SRC_DIR/$BOOT_IMG -a $BURN_IT ]; then
		MKISOFS="mkisofs -V \"$1\" -b $BOOT_IMG -c $BOOT_CAT -R $SRC_DIR"
	elif [ -f $SRC_DIR/$BOOT_IMG -a ! $BURN_IT ]; then
		MKISOFS="mkisofs -V \"$1\" -b $BOOT_IMG -c $BOOT_CAT -R $SRC_DIR -o $IMAGE_FILE"
	elif [ ! -f $SRC_DIR/$BOOT_IMG -a $BURN_IT ]; then
		print_message $ERR_DEST "WARNING: $SRC_DIR/$BOOT_IMG missing; disc will not boot"
		MKISOFS="mkisofs -V \"$1\" -R $SRC_DIR"
	elif [ ! -f $SRC_DIR/$BOOT_IMG -a ! $BURN_IT ]; then
		print_message $ERR_DEST "WARNING: $SRC_DIR/$BOOT_IMG missing; disc will not boot"
		MKISOFS="mkisofs -V \"$1\" -R $SRC_DIR -o $IMAGE_FILE"
	fi
elif [ $BURN_IT ]; then
	MKISOFS="mkisofs -V \"$1\" -R $SRC_DIR"
elif [ ! $BURN_IT ]; then
	MKISOFS="mkisofs -V \"$1\" -R $SRC_DIR -o $IMAGE_FILE"
fi

if [ $BURN_IT ]; then
	CDRECORD="cdrecord fs=6m speed=$SPEED dev=$CDRECORD_DEV -"
else
	CDRECORD=true
fi

if [ $MEDIA_TYPE = cdrw -a $BURN_IT ] && ! erase_cdrw.sh $ERR_DEST; then
	print_message $ERR_DEST "WARNING: Error erasing media"
	rm -r /tmp/be_nice
	exit 9
elif ! eval $MKISOFS 2> /dev/null | $CDRECORD 2> /dev/null; then
	print_message $ERR_DEST "WARNING: Error recording disc"
	rm -r /tmp/be_nice
	exit 8
elif [ $BURN_IT ] && ! verify_burn.sh $SRC_DIR $ERR_DEST; then
	print_message $ERR_DEST "WARNING: Filesystem corrupt"
	rm -r /tmp/be_nice
	exit 3
fi

# reincarnate_xscreensaver.sh
rm -f /tmp/be_nice
exit 0
