/*
 * 
 * 	fryit - file cdrom.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 	
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cdrom.h"
#include "globals.h"
#include "track.h"
#include "utils.h"
#include "widgets.h"

#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <fcntl.h>
#include <math.h>
#include <string>
#include <sys/fsuid.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <unistd.h>

#include <qfile.h>
#include <qregexp.h>
#include <qtextstream.h>

CDROM::CDROM(const char *v, const char *m, const char *a, const char *t,
	const char *r, int h, int b, int i, int l, int ar, const char *bd,
	const char *gd)  throw(Error) :
	Device(v,m,a,t,r,h,b,i,l,ar,bd,gd),
	rSpeed(0),
	cancelled(false) {
	
	type = cdrom;
}

void CDROM::getTOC(QList <Track> &tracks) throw(Error) {
	if (idle != getStatus()) {
		mainWin->postError("Cannot read data from CD.\n Device is being used.");
		return;
	}
	if (-2 == getCDtype())
		return;
	setStatus(other);
	
	QString out,err;
	// was -D bDev
	const char *argv[] = {Cdda2wav, "-D",devStr,"-J",0};
	readProc(out,err,argv);
	// match audio tracks
	int at = 0, nr = 0;
	QRegExp reg("[0-9]+\\.(\\s*[0-9]+:[0-9]+\\.[0-9]+)");
	while ((at = err.find(reg,at)) >= 0) {
		int min = 0, sec = 0, hun = 0;
		QString temp = err.mid(at,13);
		temp = temp.stripWhiteSpace();
		float s = 0;
		sscanf(temp,"%i.(%i:%f)",&nr,&min,&s);
		sec = (int)floor(s);
		hun = (int)rint((s - sec) * 100);
		debugmsg("audio match at %i (%s) - track %i: %i:%02i.%02i",at,temp.operator const char *(),nr,min,sec,hun);
		Track *t = new Track(nr,min,sec,hun,0,0,audio);
		CHECK_PTR(t);
		tracks.append(t);
		at += 2;
	}
	// match data tracks
	reg = "[0-9]+\\.\\[\\s*[0-9]+:[0-9]+\\.[0-9]+]";
	at = 0;
	while ((at = err.find(reg,at)) >= 0) {
		int min = 0, sec = 0, hun = 0;
		float s = 0;
		sscanf(err.mid(at),"%i.\[%i:%f]",&nr,&min,&s);
		sec = (int)floor(s);
		hun = (int)rint((s - sec) * 100);
		Track *t = new Track(nr,min,sec,hun,0,0,data1);
		CHECK_PTR(t);
		tracks.append(t);
		debugmsg("data match at %i - track %i: %i:%02i.%02i",at,nr,min,sec,hun);
		at += 2;
	}
	if (-1 == err.find("CD-Extra (CD-Plus) detected")) {
		setStatus(idle);
		return;
	}

	argv[2] = devStr;
	// readProc(out,err,argv);
	// match track names
	reg = "Track\\s*[0-9]+:\\s*'.*'";
	at = 0;
	while ((at = err.find(reg,at)) >= 0) {
		char name[128];
		if (2 == sscanf(err.mid(at),"Track %i: '%[^\n]\n",&nr,name)) {
			name[strlen(name) - 1] = '\0';
			for (Track *t = tracks.first(); t; t = tracks.next())
				if (nr == t->getNumber()) {
					debugmsg("track %i: name set to \"%s\"",nr,name);
					t->setName(name);
					break;
				}
			}
			at++;
	}
	setStatus(idle);
}

void CDROM::readDataTrack(const TrackSelection *tr,const char *filename) {
	if (idle != getStatus()) {
		mainWin->postError("Cannot read data from CD.\n Device is being used.");
		return;
	}
	setStatus(reading);
	readingAudio = false;
	thrProgress = 0;
	thrFinished = false;
	targetfile = filename;
	arg[0] = (void *) this;
	arg[1] = (void *) tr->getNumber();
	arg[2] = (void *) targetfile.operator const char *();
	arg[3] = (void *) tr->getStartOffset();
	arg[4] = (void *) tr->getEndOffset();
	debugmsg("readDataTrack(%i, %s, %i, %i)",tr->getNumber(),
		targetfile.operator const char *(),tr->getStartOffset(),
		tr->getEndOffset());
	pthread_create(&activeThr,0,readData,arg);
}

void CDROM::readToProcess(int size, Process *pr) {
	if (idle != getStatus()) {
		mainWin->postError("Cannot read data from CD.\n Device is being used.");
		return;
	}
	debugmsg("readToProcess(...)");
	setStatus(reading);
	thrProgress = 0;
	thrFinished = false;
	arg[0] = (void *) this;
	arg[1] = (void *) size;
	arg[2] = (void *) pr;
	pthread_create(&activeThr,0,readDataToProcess,arg);
}

Process *CDROM::readAudio(const TrackSelection *tr) {
	if (idle != getStatus()) {
		mainWin->postError("Cannot read data from CD.\n Device is being used.");
		return 0;
	}
	debugmsg("readAudio(...)");
	setStatus(reading);
	readingAudio = true;
	QString optT,off,len;
	optT.sprintf("-t%i",tr->getNumber());
	targetfile = DataDir;
	targetfile.append(tr->filename());
	targetfile.append(".cdr");
	const char *argv[32];
	int a = 6;
	argv[0] = Cdda2wav;
	argv[1] = "-D";
	argv[2] = devStr;
	argv[3] = optT;
	argv[4] = "-H";
	argv[5] = "-Ocdr";
	if (tr->complete()) {
		argv[a++] = "-d0";
	} else {
		int m = 0, s = 0, h = 0;
		tr->getStartTime(m,s,h);
		int sOff = (int) ceil((m * 6000 + s * 100 + h) / 100 * 75);
		debugmsg("startOffset: %i:%02i.%02i = %i",m,s,h,sOff);
		int M = 0, S = 0, H = 0;
		tr->getEndTime(M,S,H);
		debugmsg("endOffset: %i:%02i.%02i",M,S,H);
		float length = (M - m) * 60 + (S - s) + (H - h) / 100;
		debugmsg("end at %i:%02i.%02i, length %f sec.",m,s,h,length);
		off.sprintf("-o%i",sOff);
		len.sprintf("-d%f",length);
		argv[a++] = off;
		argv[a++] = len;
	}
	argv[a++] = targetfile;
	argv[a++] = 0;
	pr = new Process(argv);
	return pr;
}

int CDROM::getCDtype() throw(Error) {
	if (idle != getStatus())
		throw Error("Cannot read data from CD.\n Device is being used.");
#ifdef LINUXOS
	setStatus(other);
	int ret;
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		if ((ret = ioctl(dev,CDROM_DISC_STATUS,0)) == -1) {
			QString str;
			str.sprintf("could not get status: %s",strerror(errno));
			setStatus(idle);
			throw Error(str);
		}
	} catch (Error e) {
		setStatus(idle);
		throw e;
	}
	setStatus(idle);
	switch (ret) {
		case CDS_NO_INFO:
			errormsg("command not supported"); return 0;
		case CDS_NO_DISC:
			errormsg("no disc in drive"); return -2;
		case CDS_AUDIO:
			debugmsg("Audio CD"); return 1;
		case CDS_DATA_1:
			debugmsg("Data1 CD"); return 2;
		case CDS_DATA_2:
			debugmsg("Data2 CD"); return 3;
		case CDS_XA_2_1:
			debugmsg("XA2-1 CD"); return 4;
		case CDS_XA_2_2:
			debugmsg("XA2-2 CD"); return 5;
		case CDS_MIXED:
			debugmsg("Mixed-mode CD"); return 6;
		default:
			errormsg("unknown cd type %i",ret);
	}
	return -1;
#else
	return 0;	// command not supported
#endif
}

void CDROM::playorresume() {
	switch (getStatus()) {
		case idle:
			play(1);
			return;
		case paused:
			resume();
			return;
		default:
			QString err;
			err.sprintf("%n: cannot resume() or play()\nDevice is not paused.",
				name());
			mainWin->postError(err);
	}
}

void CDROM::play(int num, int off = 0) {
#ifdef LINUXOS	
	if (idle != getStatus() & playing != getStatus()) {
		mainWin->postError("Cannot read data from CD.\n Device is being used.");
		return;
	}
	setStatus(playing);
	infomsg("playing track %i",num);
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
  
		cdrom_tochdr hdr;
		if (ioctl(dev,CDROMREADTOCHDR,&hdr) == -1) {
			QString str;
			str.sprintf("could not get toc: %s",strerror(errno));
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		if ((num < hdr.cdth_trk0) || (num > hdr.cdth_trk1)) {
			QString str;
			str.sprintf("track number %i out of range",num);
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		cdrom_tocentry entr;
		entr.cdte_format = CDROM_MSF;
		entr.cdte_track = num;
		if (ioctl(dev,CDROMREADTOCENTRY,&entr) == -1) {
			QString str;
			str.sprintf("could not get track info: %s",strerror(errno));
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		cdrom_msf pl;
		pl.cdmsf_min0 = entr.cdte_addr.msf.minute;
		pl.cdmsf_sec0 = entr.cdte_addr.msf.second;
		pl.cdmsf_frame0 = entr.cdte_addr.msf.frame;
		pl.cdmsf_frame0 += off;
		if (pl.cdmsf_frame0 / CD_FRAMES >= 60) {
			int m = pl.cdmsf_frame0 / CD_FRAMES / 60;
			pl.cdmsf_min0 += m;
			pl.cdmsf_frame0 -= m * CD_FRAMES * 60;
		}
		if (pl.cdmsf_frame0 > CD_FRAMES) {
			int s = pl.cdmsf_frame0 / CD_FRAMES;
			pl.cdmsf_sec0 += s;
			pl.cdmsf_frame0 -= s * CD_FRAMES;
		}
		if (++entr.cdte_track > hdr.cdth_trk1)
			entr.cdte_track = CDROM_LEADOUT;
		if (ioctl(dev,CDROMREADTOCENTRY,&entr) == -1) {
			QString str;
			str.sprintf("could not get track info: %s",strerror(errno));
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		pl.cdmsf_min1 = entr.cdte_addr.msf.minute;
		pl.cdmsf_sec1 = entr.cdte_addr.msf.second;
		pl.cdmsf_frame1 = entr.cdte_addr.msf.frame;
		if (ioctl(dev,CDROMPLAYMSF,&pl) == -1) {
			QString str;
			str.sprintf("could not play track: %s",strerror(errno));
			setStatus(idle);
			mainWin->postError(str);
		}
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
#else
	mainWin->postError("Audio commands for cdplayers are only\n"
		"supported on the linux os up to now...");
#endif
}

void CDROM::pause() {
#ifdef LINUXOS
	if (playing != getStatus()) {
		mainWin->postError("Cannot pause CD.\n Device is not playing.");
		return;
	}
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		if (ioctl(dev,CDROMPAUSE,0) == -1) {
			QString str;
			str.sprintf("could not pause: %s",strerror(errno));
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		setStatus(paused);
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
#else
	mainWin->postError("Audio commands for cdplayers are only\n"
		"supported on the linux os up to now...");
#endif
}

void CDROM::resume() {
#ifdef LINUXOS
	if (getStatus() != paused) {
		mainWin->postError("Cannot resume playing CD.\n Device is not paused.");
		return;
	}
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		if (ioctl(dev,CDROMRESUME,0) == -1) {
			QString str;
			str.sprintf("could not resume playing: %s",strerror(errno));
			mainWin->postError(str);
			setStatus(idle);
			return;
		}
		setStatus(playing);
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
#else
	mainWin->postError("Audio commands for cdplayers are only\n"
		"supported on the linux os up to now...");
#endif
}

void CDROM::stop() {
#ifdef LINUXOS
	if ((playing != getStatus()) & (paused != getStatus())) {
		mainWin->postError("Cannot stop playing.\n Device is not playing.");
		return;
	}
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		if (ioctl(dev,CDROMSTOP,0) == -1) {
			QString str;
			str.sprintf("could not stop playing: %s",strerror(errno));
			mainWin->postError(str);
		}
		setStatus(idle);
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
#else
	mainWin->postError("Audio commands for cdplayers are only\n"
		"supported on the linux os up to now...");
#endif
}

void CDROM::eject() {
#ifdef LINUXOS
	if (getStatus() != idle) {
		mainWin->postError("Cannot eject CD.\n Device is being used.");
		return;
	}
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		if (ioctl(dev,CDROMEJECT,0) == -1) {
			QString err;
			err.sprintf("error ejecting cd: %s",strerror(errno));
			mainWin->postError(err);
		}
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
#else
	mainWin->postError("Ejecting of cds is only\n"
		"supported on the linux os up to now...");
#endif
}

int CDROM::getDataSize() throw(Error) {
	if (getStatus() != idle)
		throw Error("Cannot read data from CD.\n Device is being used.");
	setStatus(other);
	struct statfs st;
	QString tmp = DataDir;
	tmp += "fryit-mntpoint.XXXXXX";
	char tmpname[tmp.length() + 1];
	strcpy(tmpname,tmp);
	mktemp(tmpname);
	if (-1 == mkdir(tmpname,0)) {
		QString err = "Could not create temporary directory: ";
		err += strerror(errno);
		setStatus(idle);
		throw Error(err);
	}
	if (-1 == mount(bDev,tmpname,"iso9660",MS_RDONLY,0))  {
		QString err = "Error mounting cdrom: ";
		err += strerror(errno);
		setStatus(idle);
		throw Error(err);
	}
	if (-1 == statfs(tmpname,&st)) {
		QString err = "Could not determine filesystem size: ";
		err += strerror(errno);
		setStatus(idle);
		throw Error(err);
	}
	debugmsg("Device %s: %i blocks (%i bytes each) - total size = %i",
		name(),st.f_blocks,st.f_bsize,st.f_blocks*st.f_bsize);
	if (-1 == umount(tmpname)) {
		QString err = "Could not unmount device: ";
		err += strerror(errno);
		setStatus(idle);
		throw Error(err);
	}
	if (-1 == rmdir(tmpname)) {
		QString err = "Could not remove temporary file: ";
		err += strerror(errno);
		setStatus(idle);
		throw Error(err);
	}
	setStatus(idle);
	return st.f_bsize * st.f_blocks;
}

void CDROM::cancelAction() {
	debugmsg("CDROM::cancelAction()");
	if (readingAudio) {
		cancelled = true;
		pr->kill();
	} else if (!thrFinished) {
		debugmsg("doing pthread_cancel(activeThr)");
		pthread_cancel(activeThr);
		return;
	}
	QFile f(targetfile);
	if (f.remove())
		infomsg("%s successfully deleted",targetfile.operator const char *());
	else
		errormsg("error deleting %s: %s",targetfile.operator const char *(),
			strerror(errno));
}

void CDROM::processFinished() {
	if (readingAudio) {
		setfsuid(0);
		if (!cancelled)
			if (-1 == chown(targetfile,UserID,GroupID)) {
				QString err;
				err.sprintf("could not set file ownership: %s",strerror(errno));
				mainWin->postError(err);
			}
		cancelled = false;
		setfsuid(UserID);
		readingAudio = false;
		delete pr;
		pr = 0;
	}
	debugmsg("%s: process terminated - status set back to idle",name());
	setStatus(idle);
}

bool CDROM::mediaChanged() throw(Error) {
	if (idle != getStatus())
		throw Error("Cannot get info from CDROM.\n Device is being used.");
	setStatus(other);
	int ret;
	try {
		File dev(bDev, O_RDONLY | O_NONBLOCK, true);
		ret = ioctl(dev,CDROM_MEDIA_CHANGED,0);
		debugmsg("ioctl(dev,CDROM_MEDIA_CHANGED,0) == %i",ret);
	} catch (Error e) {
		setStatus(idle);
		throw e;
	}
	setStatus(idle);
	return ret;
}

void CDROM::msinfo(int &start, int &next) throw(Error) {
	if (idle != getStatus())
		throw Error("Cannot get info from CDROM.\n Device is being used.");
	setStatus(other);
	const char *argv[] = {Cdrecord, "-msinfo", "dev=", devStr, 0};
	QString err,out;
	readProc(out,err,argv);
	QRegExp r("[0-9]+,[0-9]+");
	int at;
	if (-1 == (at = out.find(r)))
		throw Error("could not retrive multisession info");
	sscanf(out.mid(at),"%i,%i",&start,&next);
	setStatus(idle);
}

#include "cdrom.moc"
