/*
 * 
 * 	fryit - file gui.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 	
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "cdrom.h"
#include "dialogs.h"
#include "gui.h"
#include "lowlevel.h"
#include "master.h"
#include "recorder.h"
#include "selections.h"
#include "track.h"
#include "widgets.h"

#include "globals.h"

#include <signal.h>

MainWindow::MainWindow(QApplication *a) : 
	QMainWindow(),
	qapp(a),
	file(),devices(),actions(),options(),infos(),
	errT(),
//	cdplayer(),
	confDlg(new ConfigDialog()),copyinfo(new CopyInfo()),
	mkisofsdlg(new MakeIsofsDialog()) {
	
	pthread_mutex_init(&errorMut,0);
	player = 0;
	QMenuBar *bar = menuBar();
	bar->insertItem("&File",&file);
	bar->insertItem("&Devices",&devices);
	bar->insertItem("&Actions",&actions);
	bar->insertItem("&Options",&options);
	bar->insertSeparator();
	bar->insertItem("&Infos",&infos);
	file.insertItem("&Quit",this,SLOT(quit()),CTRL+Key_Q);
	dispLog = options.insertItem("&Display Log",this,SLOT(displayLog()),CTRL+Key_L);
	options.setItemChecked(dispLog,true);
	options.insertItem("&Configuration",this,SLOT(showConfig()),CTRL+Key_O);
	options.insertItem("&Safe Options",this,SLOT(safeConfig()),CTRL+Key_S);
	actions.insertItem("&Delete Track",this,SLOT(deleteTrack()),CTRL+Key_D);
	actions.insertItem("&Play Track",this,SLOT(play()),CTRL+Key_P);
	actions.insertItem("S&top Playing",this,SLOT(stop()));
	actions.insertSeparator();
	actions.insertItem("&Copy Data CD",this,SLOT(showCopyDialog()),CTRL+Key_C);
	actions.insertItem("Create ISO9660 Image",this,SLOT(showIsofsDialog()),CTRL+Key_I);
//	actions.insertItem("&CD-Player",this,SLOT(unimplemented()));	//&cdplayer,SLOT(show()));
	infos.insertItem("About &Qt",this,SLOT(aboutQT()));
	infos.insertItem("&Version",this,SLOT(version()));
	infos.insertItem("&License",copyinfo,SLOT(show()));
	
	QFont f("Helvetica",14);
	bar->setFont(f);
	f.setPointSize(12);
	file.setFont(f);
	actions.setFont(f);
	options.setFont(f);
	infos.setFont(f);
	
	splitter = new QSplitter(QSplitter::Vertical,this);
	CHECK_PTR(splitter);
	log = new QMultiLineEdit(splitter);
	CHECK_PTR(log);
	log->setReadOnly(true);
	splitter->moveToLast(log);
	diskTrackBox = new QListView(this);
	CHECK_PTR(diskTrackBox);
	diskTrackBox->addColumn("Title");
	diskTrackBox->addColumn("Size");
	diskTrackBox->addColumn("Type");
	diskTrackBox->header()->setFont(f);
	diskTrackBox->setColumnAlignment(1,AlignCenter);
	diskTrackBox->setColumnAlignment(2,AlignCenter);
	diskTrackBox->setColumnWidth(0,150);
	diskTrackBox->setColumnWidth(1,50);
	diskTrackBox->setColumnWidth(2,50);
	removeChild(diskTrackBox);
	splitter->insertChild(diskTrackBox);
	splitter->moveToFirst(diskTrackBox);
	setCentralWidget(splitter);
	connect(diskTrackBox,SIGNAL(doubleClicked(QListViewItem *)),SLOT(play()));
	
	connect(&errT,SIGNAL(timeout()),SLOT(issueErrors()));
	errT.start(500);
	
	resize(500,350);
}

void MainWindow::configure() {
	LowLevel::cleanDatadir();
	for (Recorder *d = RecorderList.first(); d; d = RecorderList.next()) {
		debugmsg("adding recorder %s",d->name());
		RecorderMenu *m = new RecorderMenu(d);
		devices.insertItem(d->name(),m);
		writerMenus.append(m);
	}
	for (CDROM *d = CdromList.first(); d; d = CdromList.next()) {
		debugmsg("adding cdrom %s",d->name());
		CdromMenu *m = new CdromMenu(d);
		devices.insertItem(d->name(),m);
		readerMenus.append(m);
	}
//	cdplayer.configure();
}

void MainWindow::quit() {
	for (CdromMenu *cdm = readerMenus.first(); cdm; cdm = readerMenus.next()) {
		debugmsg("checking status of %s...",cdm->device()->name());
		if (idle != cdm->device()->getStatus()) {
			QString msg, cap = "Cannot quit";
			msg.sprintf("Device %s is still busy...",cdm->device()->name());
			QMessageBox::information(0,cap,msg);
			return;
		}
	}
	for (RecorderMenu *rm = writerMenus.first(); rm; rm = writerMenus.next()) {
		debugmsg("checking status of %s...",rm->device()->name());
		if (idle != rm->device()->getStatus()) {
			QString msg, cap = "Cannot quit";
			msg.sprintf("Device %s is still busy...",rm->device()->name());
			QMessageBox::information(0,cap,msg);
			return;
		}
	}
	debugmsg("quitting...");
	qapp->quit();
}

void MainWindow::version() {
	QMessageBox m("Version",
		"This is fryit Version "VERSION"\n"
		"Copyright (c) 1999\n"
		"Author: Thomas Maier-Komor\n"
		"Bugreports to: komor@eikon.tum.de\n"
		"License: GPL",QMessageBox::Information,1,0,0);
	m.setButtonText(0,"Thanks");
	QFont f("Helvetica",12);
	m.setFont(f);
	m.exec();
}

void MainWindow::aboutQT() {
	QMessageBox::aboutQt(this);
}

void MainWindow::writeLog(const char *str) {
	log->append(str);
	log->setCursorPosition(log->numLines(),0);
}

void MainWindow::displayLog() {
	if (options.isItemChecked(dispLog)) {
		splitter->removeChild(diskTrackBox);
		options.setItemChecked(dispLog,false);
		removeChild(splitter);
		insertChild(diskTrackBox);
		splitter->hide();
		setCentralWidget(diskTrackBox);
		diskTrackBox->show();
	} else {
		removeChild(diskTrackBox);
		insertChild(splitter);
		splitter->insertChild(diskTrackBox);
		splitter->moveToFirst(diskTrackBox);
		splitter->insertChild(log);
		splitter->moveToLast(log);
		setCentralWidget(splitter);
		splitter->refresh();
		diskTrackBox->move(splitter->pos());
		splitter->show();
		log->show();
		options.setItemChecked(dispLog,true);
	}
}

void MainWindow::showCopyDialog() {
	CopySelectionDialog *copyDlg = new CopySelectionDialog();
	CHECK_PTR(copyDlg);
	copyDlg->show();
}

void MainWindow::showIsofsDialog() {
	mkisofsdlg->show();
}

void MainWindow::showConfig() {
	confDlg->show();
}

void MainWindow::safeConfig() {
	LowLevel::writeConfig();
}

void MainWindow::resizeEvent(QResizeEvent *) {
	if (options.isItemChecked(dispLog))
		diskTrackBox->move(splitter->pos());
	diskTrackBox->repaint();
}

void MainWindow::updateView() {
	diskTracks.clear();
	diskTrackBox->clear();
	QList <Track> tracks;
	LowLevel::getDiscTracks(tracks);
	for (Track *tr = tracks.first(); tr; tr = tracks.next()) {
		TrackItem *t = new TrackItem(tr,diskTrackBox);
		diskTracks.append(t);
	}
}

void MainWindow::deleteTrack() {
	TrackItem *t = (TrackItem *) diskTrackBox->currentItem();
	if (!t)
		return;
	LowLevel::deleteTrack(t->tr);
	delete t;
	updateView();
}

void MainWindow::play() {
	TrackItem *ti = (TrackItem *) diskTrackBox->currentItem();
	if (!ti)
		return;
	const Track *t = ti->tr;
	if (player) {
		debugmsg("playDiskTrack(): terminating old process");
		player->terminate();
		delete player;
		player = 0;
	}
	try {
		player = LowLevel::playDiskTrack(t);
	}
	catch (Error e) {
		QMessageBox::critical(0,"Error",e);
	}
	connect(player,SIGNAL(quit(int)),SLOT(playerStopped(int)));
}

void MainWindow::stop() {
	if (player) {
		player->terminate();
		delete player;
		player = 0;
	}
}

void MainWindow::pause() {
	if (player) {
		debugmsg("pauseing player");
		player->stop();
	}
}

void MainWindow::resume() {
	if (player) {
		debugmsg("resume player");
		player->cont();
	}
}

void MainWindow::playerStopped(int r) {
	infomsg("player exited with code %i",r);
	delete player;
	player = 0;
}

void MainWindow::unimplemented() {
	QMessageBox::critical(0,
		"unimplemented...",
		"The requested feature is not\n"
		"implemented up to now.\n"
		"Wait for the next release.");
}

void MainWindow::postError(const QString &err) {
	debugmsg("postError(\"%s\") - error logged",err.operator const char *());
	pthread_mutex_lock(&errorMut);
	errors.append(err);
	pthread_mutex_unlock(&errorMut);
}

void MainWindow::issueErrors() {
	QString cap = "Error";
	pthread_mutex_lock(&errorMut);
	while (errors.count()) {
		QString msg = errors.getFirst();
		errors.removeFirst();
		pthread_mutex_unlock(&errorMut);
		QMessageBox::critical(0,cap,msg);
		pthread_mutex_lock(&errorMut);
	}
	pthread_mutex_unlock(&errorMut);
}

CdromMenu::CdromMenu(CDROM *d) :
	QPopupMenu(),
	cdr(d),
	rsd(new ReadSelectionDialog(d)) {
	
	setItemEnabled(insertItem(d->name()),false);
	insertSeparator();
	insertItem("Eject CD",cdr,SLOT(eject()));
	play = insertItem("Play CD",cdr,SLOT(playorresume()));
	pause = insertItem("Pause CD",cdr,SLOT(pause()));
	setItemEnabled(pause,false);
	stop = insertItem("Stop CD",cdr,SLOT(stop()));
	setItemEnabled(stop,false);
	insertItem("Read Tracks",rsd,SLOT(showDialog()));
	connect(d,SIGNAL(isOperating(status)),SLOT(enable(status)));
}

void CdromMenu::enable(status s) {
	switch (s) {
		case idle :
			for (int i = 2; -1 != idAt(i); i++)
				if ((stop != idAt(i)) & (pause != idAt(i)))
					setItemEnabled(idAt(i),true);
				else
					setItemEnabled(idAt(i),false);
			break;
		case playing :
			for (int i = 2; -1 != idAt(i); i++)
				if ((stop != idAt(i)) & (pause != idAt(i)))
					setItemEnabled(idAt(i),false);
				else
					setItemEnabled(idAt(i),true);
			break;
		case paused :
			for (int i = 2; -1 != idAt(i); i++)
				if ((stop != idAt(i)) & (play != idAt(i)))
					setItemEnabled(idAt(i),false);
				else
					setItemEnabled(idAt(i),true);
			break;
		case other :
		case reading :
		case writing :
			for (int i = 2; -1 != idAt(i); i++)
				setItemEnabled(idAt(i),false);
			break;
		default:
			warningmsg("*Menu got unknown status signal %i. This is a bug!",s);
	}
}

RecorderMenu::RecorderMenu(Recorder *rec) :
	CdromMenu(rec),
	wsd(new WriteSelectionDialog(rec)),
	dcd(new DeviceConfigDialog(rec)) {
	
	insertItem("Write Tracks",wsd,SLOT(show()));
	insertItem("Configure",dcd,SLOT(show()));
	insertItem("Eject CD",rec,SLOT(eject()));
}

#include "gui.moc"
