/*
 * 
 * 	fryit - file lowlevel.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 	
 */

#include "config.h"

#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <sys/fsuid.h>
#include <sys/stat.h>
#include <unistd.h>

#include <qdatastream.h>
#include <qdir.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include <qtextstream.h>

#include "cdrom.h"
#include "recorder.h"
#include "track.h"
#include "lowlevel.h"
#include "widgets.h"
#include "globals.h"


#ifdef NAMESPACES
namespace LowLevel {
#endif
	int
		LowLevel::majVer, LowLevel::minVer, LowLevel::patVer;
	bool
		LowLevel::superuser;
	QString
		LowLevel::cdrecord, LowLevel::cdda2wav, LowLevel::mkisofs,
		LowLevel::playcdr, LowLevel::dataDir;
	QList <Device> 
		LowLevel::deviceList;
	QList <CDROM>
		LowLevel::cdromList;
	QList <Recorder>
		LowLevel::recorderList;
	QStringList
		LowLevel::drivers,LowLevel::driverDescr;
	pthread_mutex_t
		LowLevel::idMut;
	QValueList <int>
		LowLevel::audioIDs,LowLevel::dataIDs;
	uid_t
		LowLevel::userID;
	gid_t
		LowLevel::groupID;

#ifdef NAMESPACES
	void scanSCSIDevices();
	int checkRecorder(int,int,int);
	void setupDriverList();
	void initPathes();
	void setupLists();
	void setupIDlists();
	Device *getDevice(QString &dev, int bus, int id, int lun);
	bool boolvalue(const QString &);
	void setOption(const QString &, const QString &);
	void setRootOption(const QString &, const QString &);
	void readGlobalConfig();
	void readLocalConfig();
}
#endif

int LowLevel::init() {
	pthread_mutex_init(&idMut,0);
	
	dataDir = "~/fryit";
	initPathes();
	userID = getuid();
	groupID = getgid();
	
	if (userID == 0)
		superuser = true;
	else if (setuid(0) == -1) {
		errormsg("cannot set uid to root: %s",strerror(errno));
		QMessageBox m("Incorrect installation!",
			"Cannot setuid root.\n"
			"Ask root to do a chmod +s fryit and chown root fryit.\n"
			"Fryit will probably not work without that...",
			QMessageBox::Critical,1,0,0);
		m.setButtonText(0,"Thanks");
		QFont f("Helvetica",12);
		m.setFont(f);
		m.exec();
	}
	setfsuid(userID);
	setfsgid(groupID);
	
	setupLists();
	readGlobalConfig();
	scanSCSIDevices();
	readLocalConfig();
	QString out,err;
	const char *argv[3] = {cdrecord,"-version",0};
	readProc(out,err,argv);
	if (err != "") {
		errormsg("cannot get cdrecord version!");
		errormsg("\"cdrecord -version\" returned: %s",err.operator const char *());
		errormsg("ignoring error, although probably fatal...");
	} else if (sscanf(out.operator const char *(),"Cdrecord release %i.%i.%i",
		&majVer,&minVer,&patVer) < 2) {
		errormsg("LowLevel::init(): could not analyze version info");
		errormsg("ignoring error...");
	} else
		infomsg("cdrecord version %i.%i.%i found.",majVer,minVer,patVer);
	return 0;
}

void LowLevel::setupLists() {
	// setup driver list
	QString out,err,tmp;
	const char *argv[] = { cdrecord, "-driver=help", 0 };
	readProc(out,err,argv);
	QTextStream ts(err,IO_ReadOnly);
	ts.readLine();	// ignore first line
	tmp = "Autodetect driver";
	driverDescr.append(tmp);
	tmp = "";
	drivers.append(tmp);
	while (!ts.atEnd()) {
		ts >> tmp;
		tmp.prepend("-driver=");
		drivers.append(tmp);
		tmp = ts.readLine();
		tmp = tmp.simplifyWhiteSpace();
		driverDescr.append(tmp);
	}

	setupIDlists();
}

void LowLevel::setupIDlists() {
	audioIDs.clear();
	dataIDs.clear();
	QDir d(dataDir);
	d.setNameFilter("audio*.cdr");
	for (unsigned c = 0; c < d.count(); c++) {
		int id;
		if (1 == sscanf(d[c],"audio%i.cdr",&id))
			if (!audioIDs.contains(id))
				audioIDs.append(id);
			else
				warningmsg("Multiple existence of audiofiles with id %i.\n"
					"Will only access audio%i.cdr.",id);
	}
	d.setNameFilter("data*.cdr");
	for (unsigned c = 0; c < d.count(); c++) {
		int id;
		if (1 == sscanf(d[c],"data%i.cdr",&id)) {
			if (!dataIDs.contains(id))
				dataIDs.append(id);
			else
				warningmsg("Multiple existence of datafiles with id %i.\n"
					"Will only access data%i.cdr.",id);
		}
	}
	
}

int LowLevel::checkRecorder(int host, int id, int lun) {
	QString out,err,dev;
	dev.sprintf("dev=%i,%i,%i",host,id,lun);
	const char *argv[4] = {cdrecord,"-checkdrive",dev,0};
	return readProc(out,err,argv);
}

void LowLevel::scanSCSIDevices() {
#ifdef PROCSCAN
	QFile in("/proc/scsi/scsi");
	if (!in.open(IO_ReadOnly)) {
		errormsg("could not open /proc/scsi/scsi: %s",strerror(errno));
		return;
	}
	QTextStream ts(&in);
	QString tmp = ts.readLine();
	if (strcmp(tmp,"Attached devices: ") != 0) {
		fatalmsg("format of /proc/scsi/scsi seems to have changed");
		return;
	}
	
	int sg = 0, scd = 0;
	char sd = 'a';
	while (!in.atEnd()) {
		debugmsg("in at %i/%i",in.at(),in.size());
		QString adapter,vendor,model,rev,type,bDev,gDev;
		int host,bus,id,lun,ansirev;
		ts >> tmp >> adapter;
		if (adapter.isEmpty()) {
			debugmsg("adapter is empty");
			break;
		}
		debugmsg("adapter: %s",adapter.operator const char *());
		ts >> tmp >> bus >> tmp >> id >> tmp >> lun;
		ts >> tmp >> vendor;
		debugmsg("vendor: %s",vendor.operator const char *());
		ts >> tmp;
		while (strcmp(tmp,"Model:") != 0) {
			vendor += ' ';
			vendor += tmp;
			ts >> tmp;
		}
		ts >> model;
		ts >> tmp;
		while (strcmp(tmp,"Rev:") != 0) {
			model += ' ';
			model += tmp;
			ts >> tmp;
		}
		debugmsg("model: %s",model.operator const char *());
		ts >> rev;
		ts >> tmp >> type;
		ts >> tmp;
		while (strcmp(tmp,"ANSI") != 0) {
			type += ' ';
			type += tmp;
			ts >> tmp;
		}
		ts >> tmp >> tmp >> ansirev;
		debugmsg("ansi rev: %i",ansirev);
		tmp = adapter;
		tmp.remove(0,4);
		host = tmp.toInt();
		gDev.sprintf("/dev/sg%i",sg++);
		
		Device *d = 0;
		if (!strcmp(type,"CD-ROM")) {
			bDev.sprintf("/dev/scd%i",scd++);
			if (0 == checkRecorder(host,bus*16+id,lun)) {
				d = new Recorder(vendor,model,adapter,type,rev,host,bus,id,lun,ansirev,bDev,gDev);
				infomsg("adding recorder %s to device list", d->name());
				recorderList.append(d);
			} else {
				d = new CDROM(vendor,model,adapter,type,rev,host,bus,id,lun,ansirev,bDev,gDev);
				infomsg("adding cdrom %s to device list", d->name());
				cdromList.append(d);
			}
		} else if (!strcmp(type,"Direct-Access")) {
			bDev = "/dev/sd";
			bDev += sd++;
			d = new Device(vendor,model,adapter,type,rev,host,bus,id,lun,ansirev,bDev,gDev);
			infomsg("adding harddisk %s to device list", d->name());
		} else if (!strcmp(type,"Optical Device")) {
			bDev = "/dev/sd";
			bDev += sd++;
			d = new Device(vendor,model,adapter,type,rev,host,bus,id,lun,ansirev,bDev,gDev);
			infomsg("adding harddisk %s to device list", d->name());
		} else
			assert(0);
		deviceList.append(d);
	}
	return;
#else
	const char *argv[] = {cdrecord, "-scanbus", 0};
	QString out,err;
	readProc(out,err,argv);
	QTextIStream ts(&out);
	QRegExp reg("[0-9]+) '.+' '.+' '.+' .+");
	int sg = 0, scd = 0, bus = 0;
	char sd = 'a';
	QString adapter = "scsi0";
	while (!ts.atEnd()) {
		int at = 0;
		QString line = ts.readLine();
		if (1 == sscanf(line,"scsibus%i",&bus)) {
			debugmsg("scanSCSIDevices(): bus set to %i",bus);
			adapter.sprintf("scsi%i",bus);
			continue;
		} else if (-1 == (at = line.find(reg))) 
			continue;
		int id = 0;
		char vendor[9],model[17],rev[5],type[32];
		debugmsg("scanSCSIDevices(): parsing \"%s\"",line.mid(at).operator const char *());
		sscanf(line.mid(at),"%i) '%[^']' '%[^']' '%[^']' %[^\n]\n",
			&id,vendor,model,rev,type);
		if (id > 15)
			id %= 100;
		QString bDev,gDev;
		Device *dev = 0;
		gDev = "/dev/sg";
		gDev += sg++;
		if (!strcmp(type,"Disk")) {
			bDev = "/dev/sd";
			bDev += sd++;
			dev = new Device(vendor,model,adapter,type,rev,bus,0,id,0,0,bDev,gDev);
		} else if (!strcmp(type,"Removable CD-ROM")) {
			bDev.sprintf("/dev/scd%i",scd++);
			setfsuid(0);
			if (0 == checkRecorder(bus,id,0)) {
				Recorder *rec;
				try { rec = new Recorder(vendor,model,adapter,type,rev,bus,0,id,0,0,bDev,gDev); }
				catch (Error e) {mainWin->postError(e); continue;}
				dev = rec;
				infomsg("adding recorder %s to device list", dev->name());
				recorderList.append(rec);
			} else {
				CDROM *cdr;
				try {cdr = new CDROM(vendor,model,adapter,type,rev,bus,0,id,0,0,bDev,gDev);}
				catch (Error e) {mainWin->postError(e); continue;}
				dev = cdr;
				infomsg("adding cdrom %s to device list", dev->name());
				cdromList.append(cdr);
			}
			setfsuid(UserID);
		} else if (0 == strcmp(type,""))
			assert(0);
		deviceList.append(dev);
	}
	return;
#endif
}

int LowLevel::newTrackID(bool audio = true) {
	if (audio) {
		for (int c = 1; ; c++) {
			debugmsg("checking if audioid %i is already in use",c);
			if (!audioIDs.contains(c)) {
				audioIDs.append(c);
				debugmsg("created new audio ID %i",c);
				return c;
			}
		}
	} else {
		for (int c = 1; ; c++) {
			debugmsg("checking if dataid %i is already in use",c);
			if (!dataIDs.contains(c)) {
				dataIDs.append(c);
				debugmsg("created new data ID %i",c);
				return c;
			}
		}
	}
	assert(0);
	return 0;
}

void LowLevel::readGlobalConfig() {
	debugmsg("reading global config file...");
	QFile cfgf(ROOTCONFIGFILE);
	if (!cfgf.open(IO_ReadOnly)) {
		infomsg("could not open config file: %s",strerror(errno));
		if (!Superuser) {
			QMessageBox::critical(0,"Incorrect installation!",
					      "For security reasons fryit will "
					      "not start without global config file!",
					      "Sorry");
			exit(-1);
			return;
		}
	} else {
		QTextStream cfg(&cfgf);
		int linenr = 0;
		while (!cfg.atEnd()) {
			QString line = cfg.readLine();
			linenr++;
			debugmsg("parsing line %i: \"%s\"",linenr,line.operator const char *());
			if (('#' == line[0]) | line.isEmpty()) {	// ignore comments and empty lines
				debugmsg("comment line");
				continue;
			}
			int set = line.find(':');
			if (set == -1) {
				set = line.find('=');
				if (set == -1) {
					errormsg("error parsing config file in line %i: ignoring line",linenr);
					continue;
				}
			}
			QString key = line.left(set);
			key = key.simplifyWhiteSpace();
			QString value = line.mid(set + 1);
			int comm;
			if (-1 != (comm = value.find('#')))	// remove comment at end of line
				value = value.left(comm);
			value = value.simplifyWhiteSpace();
			setRootOption(key,value);
		}
	}
}

void LowLevel::readLocalConfig() {
	debugmsg("reading local config file...");
	QFile cfgf(QDir::home().absPath() + "/.fryit.conf");
	QTextStream cfg(&cfgf);
	if (!cfgf.open(IO_ReadOnly)) {
		infomsg("could not open personal config file: %s",strerror(errno));
		return;
	}
	int linenr = 0;
	while (!cfg.atEnd()) {
		QString line = cfg.readLine();
		linenr++;
		debugmsg("parsing line %i: \"%s\"",linenr,line.operator const char *());
		if (('#' == line[0]) | line.isEmpty()) {	// ignore comments and empty lines
			debugmsg("comment line");
			continue;
		}
		int set = line.find(':');
		if (set == -1) {
			set = line.find('=');
			if (set == -1) {
				errormsg("error parsing config file in line %i: ignoring line",linenr);
				continue;
			}
		}
		QString key = line.left(set);
		key = key.simplifyWhiteSpace();
		QString value = line.mid(set + 1);
		int comm;
		if (-1 != (comm = value.find('#')))	// remove comment at end of line
			value = value.left(comm);
		value = value.simplifyWhiteSpace();
		setOption(key,value);
	}
}

void LowLevel::writeConfig() {
	QString str = QDir::home().absPath() + "/.fryit.conf";
	QFile cfgf(str);
	if (!cfgf.open(IO_WriteOnly)) {
		QString err = "Error writing config file:\n";
		err += strerror(errno);
		mainWin->postError(err);
		return;
	}
	QTextStream cfg(&cfgf);
	cfg << "#\n# config file for fryit\n#\n\n";
	cfg << "verboselevel = " << verboseLvl << "\n";
	cfg << "datadir = " << dataDir << "\n\n";
	for (Recorder *r = recorderList.first(); r; r = recorderList.next()) {
		cfg << "device : " << r->name() << "\n";
		cfg << "at : " << r->host() << "," << r->id() << "," << r->lun() << "\n";
		if (r->getDriver())
			cfg << "driver: " << drivers[r->getDriver()].mid(8) << "\n";
		cfg << "speed: " << r->getWriteSpeed() << "\n";
		cfg << "fifo: " << r->getFifo() << "\n";
		cfg << "eject: " << ((r->getEject()) ? "yes" : "no") << "\n";
		cfg << "simulate: " << ((r->getSimulate()) ? "yes" : "no") << "\n";
		cfg << "multisession: " << ((r->getMultisession()) ? "yes" : "no") << "\n";
		cfg << "\n";
	}
	for (CDROM *c = cdromList.first(); c; c = cdromList.next()) {
		if (c->getReadSpeed()) {
			cfg << "device: " << c->name() << "\n";
			cfg << "at: " << c->host() << "," << c->id() << "," << c->lun() << "\n";
			cfg << "speed: " << c->getReadSpeed() << "\n";
			cfg << "\n";
		}
	}
	//mainWin->writeConfig(cfg);

	// write superuser config
	if (Superuser) {
		debugmsg("writing superuser config");
		cfgf.close();
		cfgf.setName(ROOTCONFIGFILE);
		if (!cfgf.open(IO_ReadWrite)) {
			QString err = "Could not write superuser config: ";
			err += strerror(errno);
			mainWin->postError(err);
			return;
		}
		cfg << "cdrecord = " << cdrecord << "\n";
		cfg << "cdda2wav = " << cdda2wav << "\n";
		cfg << "mkisofs = " << mkisofs << "\n";
		cfg << "playcdr = " << playcdr << "\n";
	}
	infomsg("config written to ~/.fryit.conf");
}

void LowLevel::writeTrackInfo(const TrackSelection *t) {
	debugmsg("writing track info...");
	QString str;
	str = dataDir + t->filename() + ".inf";
	QFile inf(str);
	if (!inf.open(IO_WriteOnly)) {
		QString str;
		str.sprintf("Could not write track information file!\nReason: %s",
			strerror(errno));
		QMessageBox::critical(0,"Error",
			str,
			"Oh no!");
		return;
	}
	QTextStream ds(&inf);
	ds << "Name: " << t->getName() << "\n";
	trackType ty;
	int i0,i1,i2;
	if (audio == t->getTrackType()) {
		t->getSelectedTime(i0,i1,i2);
		str.sprintf("Time: %i:%02i.%02i\n",i0,i1,i2);
	} else {
		str.sprintf("Length: %li\n", 
			(t->getEndOffset() - t->getStartOffset()) * 1024);
		debugmsg("data track with length %li byte",(t->getEndOffset() - t->getStartOffset()) * 1024);
	}
	ds << str;
	t->getOptions(i0,i1,ty);
	ds << "adr: " << str.setNum(i0) << "\n";
	ds << "control: " << str.setNum(i1) << "\n";
	ds << "mode: " << str.setNum((int) ty) << "\n";
	inf.close();
	return;
}

void LowLevel::getDiscTracks(QList <Track> &tr) {
	QDir d(dataDir,"*.inf");
	for (unsigned c = 0; c < d.count(); c++) {
		QString fn = dataDir;
		fn.append(d[c]);
		debugmsg("reading info file %s",fn.operator const char *());
		QFile f(fn);
		if (!f.open(IO_ReadOnly)) {
			errormsg("could not open file %s",d[c].operator const char *());
			continue;
		}
		QTextStream ts(&f);
		bool audio;
		int id = 0;
		if (0 == strncmp(d[c],"audio",5)) {
			audio = true;
			sscanf(d[c],"audio%i",&id);
		} else if (0 == strncmp(d[c],"data",4)) {
			audio = false;
			sscanf(d[c],"data%i",&id);
		} else
			continue;
		QString name;
		int m,s,h,a = -1,c = -1;
		long le = -1;
		trackType mo = unknown;
		for (QString line = ts.readLine(); !(line.isEmpty()&&ts.atEnd()) ; line = ts.readLine()) {
			line = line.simplifyWhiteSpace();
			if (line[0] == '#') // comment line
				continue;
			if (line.find("name",0,false) == 0) {
				line.remove(0,5);
				if (line[0] == ':')
					line.remove(0,1);
				name = line.simplifyWhiteSpace();
			} else
			if (line.find("time",0,false) == 0) {
				line.remove(0,5);
				if (line[0] == ':')
					line.remove(0,1);
				sscanf(line.operator const char *(),"%d:%d.%d",&m,&s,&h);
			} else 
			if (line.find("length",0,false) == 0) {
				line.remove(0,7);
				if (line[0] == ':')
					line.remove(0,1);
				sscanf(line.operator const char *(),"%ld",&le);
			} else 
			if (line.find("mode",0,false) == 0) {
				line.remove(0,5);
				if (line[0] == ':')
					line.remove(0,1);
				mo = (trackType) line.toInt(0);
			} else 
			if (line.find("control",0,false) == 0) {
				line.remove(0,8);
				if (line[0] == ':')
					line.remove(0,1);
				c = line.toInt(0);
			} else 
			if (line.find("adr",0,false) == 0) {
				line.remove(0,4);
				if (line[0] == ':')
					line.remove(0,1);
				a = line.toInt(0);
			}
		}
		f.close();
		fn.replace(fn.length() - 3,3,"cdr");
		debugmsg("checking file %s",fn.operator const char *());
		f.setName(fn);
		f.open(IO_ReadOnly);
		if (-1 == le) {
			le = (int) rint( (float) (m * 6000 + s * 100 + h) / 100 * 75 * 2352);
			debugmsg("got length inf in time (%i:%02i.%02i) = %i byte)",m,s,h,le);
		}
		long l = f.size();
		if (l !=  le) {
			int tmp = l / 2352;
			tmp = (int) rint(tmp / 75 * 100);
			m = tmp / 6000;
			tmp %= 6000;
			s = tmp / 100;
			tmp %= 100;
			h = tmp;
			warningmsg("info file for track \"%s\" delivers incorrect"
				" size %li - using %li of datafile\ncorresponding time is %i:%02i.%02i",
				name.operator const char *(),le,l,m,s,h);
			le = l;
		}
		Track *t = new Track(0,le,a,c,mo,name,id);
		CHECK_PTR(t);
		tr.append(t);
		f.close();
	}
}

void LowLevel::cleanDatadir() {
	debugmsg("cleanDatadir()");
	bool first = true;
	QDir d(dataDir,"*.inf");
	debugmsg("%i .inf files",d.count());
	for (unsigned c = 0; c < d.count(); c++) {
		const char *basename = d[c];
		QString filename = dataDir + basename;
		filename.replace(filename.find(".inf"),4,".cdr");
		debugmsg("checking file %s",filename.operator const char *());
		QFile f(filename);
		if (((!f.exists()) || (0 == f.size())) 
			&& ((strncmp(basename,"audio",5) == 0) 
			|| (strncmp(basename,"data",4) == 0))) {
			if (f.exists())
				f.remove();
			if (first) {
				first = false;
				infomsg("cleaning up data directory:");
			}
			filename = dataDir + basename;
			f.setName(filename);
			debugmsg("%s does not have a .cdr file",basename);
			infomsg("	removing info file %s",basename);
			if (!f.remove())
				errormsg("could not remove file %s: %s",basename,strerror(errno));
		}
	}
	mainWin->updateView();
}

Process *LowLevel::playDiskTrack(const Track *t, int off = 0) {
	QString file = dataDir;
	file += t->filename();
	file += ".cdr";
	if (off) {
		QString offset;
		offset.setNum(off);
		const char *argv[] = {LowLevel::playcdr,"-f",offset,file,0};
		return new Process(argv);
	} else {
		const char *argv[] = {LowLevel::playcdr,file,0};
		return new Process(argv);
	}
}

void LowLevel::deleteTrack(const Track *tr) {
	debugmsg("deleting file %s",tr->filename());
	QString fn = dataDir;
	fn += tr->filename();
	fn += ".inf";
	if (QFile::remove(fn))
		infomsg("removed file %s",fn.operator const char *());
	else
		errormsg("could not remove file %s: %s",
			fn.operator const char *(),strerror(errno));
	fn = dataDir;
	fn += tr->filename();
	fn += ".cdr";
	if (QFile::remove(fn))
		infomsg("removed file %s",fn.operator const char *());
	else
		errormsg("could not remove file %s: %s",
			fn.operator const char *(),strerror(errno));
}

void LowLevel::setDataDir(const char *d) {
	if (dataDir == d)
		return;
	dataDir = d;
	if (dataDir[dataDir.length() - 1] != '/')
		dataDir.append("/");
	QDir dir(dataDir);
	if (!dir.exists()) {
		errormsg("directory \"%s\" does not exist. datadir unchanged.");
		return;
	}
	debugmsg("dataDir set to %s",dataDir.operator const char *());
	DataDir = dataDir;
	setupIDlists();
	mainWin->updateView();
}

void LowLevel::initPathes() {
	bool cdr = false, mkiso = false, cdda = false, play = false;
	QString dirs = getenv("PATH");
	debugmsg("dirs = %s",dirs.operator const char *());
	while (!dirs.isEmpty()) {
		QString dir,name;
		int pos = dirs.find(':');
		dir = dirs;
		if (-1 == pos)
			dirs = "";
		else {
			dir = dirs.left(pos);
			dirs = dirs.mid(pos+1);
		}
		dir += '/';
		debugmsg("checking directory %s",dir.operator const char *());
		if (!cdr) {
			name = "cdrecord";
			QFileInfo i(dir,name);
			if (i.exists() & i.isExecutable()) {
				cdr = true;
				cdrecord = dir+name;
				Cdrecord = cdrecord;
				infomsg("cdrecord found in %s",dir.operator const char *());
			}
		}
		if (!mkiso) {
			name = "mkisofs";
			QFileInfo i(dir,name);
			if (i.exists() & i.isExecutable()) {
				mkiso = true;
				mkisofs = dir+name;
				Mkisofs = mkisofs;
				infomsg("mkisofs found in %s",dir.operator const char *());
			}
		}
		if (!cdda) {
			name = "cdda2wav";
			QFileInfo i(dir,name);
			if (i.exists() & i.isExecutable()) {
				cdda = true;
				cdda2wav = dir+name;
				Cdda2wav = cdda2wav;
				infomsg("cdda2wav found in %s",dir.operator const char *());
			}
		}
		if (!play) {
			name = "playcdr";
			QFileInfo i(dir,name);
			if (i.exists() & i.isExecutable()) {
				play = true;
				playcdr = dir+name;
				Playcdr = playcdr;
				infomsg("playcdr found in %s",dir.operator const char *());
			}
		}
		
	}
	
	// datadir
	if ('~' == dataDir[0]) {
		dataDir = dataDir.mid(1);
		dataDir.prepend(QDir::home().absPath());
		if ('/' != dataDir[dataDir.length()])
			dataDir += '/';
		DataDir = dataDir;
	}
	QFileInfo f(dataDir);
	if (!(f.exists() & f.isReadable() & f.isWritable()))
		if (f.exists())
			warningmsg("datadir \"%s\" has incorrect access rights...",
				dataDir.operator const char *());
		else {
			infomsg("creating data directory \"%s\"",dataDir.operator const char *());
			QDir d;
			if (-1 == mkdir(dataDir,0755)) {
				QString err = "Error creating data directory \"";
				err += dataDir;
				err += "\": ";
				err += strerror(errno);
				mainWin->postError(err);
			}
		}
	QDir::setCurrent(dataDir);
}

void LowLevel::setMkisofs(const char *p) {
	if (mkisofs == p)
		return;
	QFileInfo i(p);
	if (!i.isExecutable()) {
		errormsg("\"%s\" is invalid or not executable. Value unmodified.",p);
		return;
	}
	infomsg("using \"%s\" for mkisofs",p);
	mkisofs = p;
	Mkisofs = mkisofs;
}

void LowLevel::setCdrecord(const char *p) {
	if (cdrecord == p)
		return;
	QFileInfo i(p);
	if (!i.isExecutable()) {
		errormsg("\"%s\" is invalid or not executable. Value unmodified.",p);
		return;
	}
	infomsg("using \"%s\" for cdrecord",p);
	cdrecord = p;
	Cdrecord = cdrecord;
}

void LowLevel::setPlaycdr(const char *p) {
	if (playcdr == p)
		return;
	QFileInfo i(p);
	if (!i.isExecutable()) {
		errormsg("\"%s\" is invalid or not executable. Value unmodified.",p);
		return;
	}
	infomsg("using \"%s\" for playcdr",p);
	playcdr = p;
	Playcdr = playcdr;
}

void LowLevel::setCdda2wav(const char *p) {
	if (cdda2wav == p)
		return;
	QFileInfo i(p);
	if (!i.isExecutable()) {
		errormsg("\"%s\" is invalid or not executable. Value unmodified.",p);
		return;
	}
	infomsg("using \"%s\" for cdda2wav",p);
	cdda2wav = p;
	Cdda2wav = cdda2wav;
}

Device *LowLevel::getDevice(QString &dev, int bus, int id, int lun) {
	debugmsg("searching device \"%s\"",dev.operator const char *());
	for (Device *d = deviceList.first(); d; d = deviceList.next()) {
		debugmsg("checking device \"%s\" at %i,%i,%i",d->name(),
			d->bus(),d->id(),d->lun());
		if (!strcmp(d->name(),dev)) {
			if (bus == -1)
				return d;
			else if ((d->host() == bus) & (d->id() == id) & (d->lun() == lun))
				return d;
		}
	}
	return 0;
}

bool LowLevel::boolvalue(const QString &value) {
	if ((!strcasecmp("yes",value)) | (!strcasecmp("true",value)) | (!strcasecmp("on",value)))
		return true;
	if ((!strcasecmp("no",value)) | (!strcasecmp("false",value)) | (!strcasecmp("off",value)))
		return false;
	warningmsg("unexpected bool value \"%s\" - assuming false",value.operator const char *());
	return false;
}

void LowLevel::setRootOption(const QString &key, const QString &value) {
	debugmsg("setRootOption(\"%s\",\"%s\")",
		key.operator const char *(),
		value.operator const char *());
	if ("cdrecord" == key)
		setCdrecord(value);
	else if ("cdda2wav" == key)
		setCdda2wav(value);
	else if ("playcdr" == key)
		setPlaycdr(value);
	else if ("mkisofs" == key)
		setMkisofs(value);
	else
		errormsg("unexpected key \"%s\" with value \"%s\" in root config file: ingnored",
			key.operator const char *(),
			value.operator const char *());
}

void LowLevel::setOption(const QString &key, const QString &value) {
	static QString dev = "", idStr = "";
	static int bus = -1, id = -1, lun = -1;

	debugmsg("setOption(\"%s\",\"%s\")",
		key.operator const char *(),
		value.operator const char *());
	if ("verboselevel" == key) {
		bool ok = false;
		int v = value.toInt(&ok);
		if (ok)
			verboseLvl = v;
	} else if ("datadir" == key)
		setDataDir(value);
	else if (!strcasecmp("device",key)) {
		dev = value;
		bus = id = lun = -1;
		idStr = "";
	} else if (!strcasecmp("at",key)) {
		bus = id = lun = 0;
		idStr = value;
		sscanf(value,"%i,%i,%i",&bus,&id,&lun);
	} else if (!dev) {
		errormsg("Cannot set parameter for device without a device!");
		return;
	} else if (!strcasecmp("fifo",key)) {
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		if (recorder != d->getType()) {
			errormsg("cannot set eject flag: \"%s\" is not a recorder",
				dev.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		bool ok = false;
		int f = value.toInt(&ok);
		if (ok)
			r->setFifo(f);
		else
			errormsg("unexpected value \"%s\" for key \"%s\"",
				value.operator const char *(),
				key.operator const char *());
	} else if (!strcasecmp("eject",key)) {
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		if (recorder != d->getType()) {
			errormsg("cannot set eject flag: \"%s\" is not a recorder",
				dev.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		if ((!strcasecmp("no",value)) | (!strcasecmp("false",value))) {
			infomsg("setting eject to false for \"%s\"",
				dev.operator const char *());
			r->setEject(false);
		} else if ((!strcasecmp("yes",value)) | (!strcasecmp("true",value))) {
			infomsg("setting eject to true for \"%s\"",
				dev.operator const char *());
			r->setEject(true);
		} else
			errormsg("unexpected value \"%s\" for key \"%s\"",
				value.operator const char *(),
				key.operator const char *());
	} else if (!strcasecmp("simul",key) | !strcasecmp("simulate",key)) {
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		if (recorder != d->getType()) {
			errormsg("cannot set simulation flag: \"%s\" is not a recorder",
				dev.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		if ((!strcasecmp("no",value)) | (!strcasecmp("false",value))) {
			infomsg("setting simulation to false for \"%s\"",
				dev.operator const char *());
			r->setSimulate(false);
		} else if ((!strcasecmp("yes",value)) | (!strcasecmp("true",value))) {
			infomsg("setting simulation to true for \"%s\"",
				dev.operator const char *());
			r->setSimulate(true);
		} else
			errormsg("unexpected value \"%s\" for key \"%s\"",
				value.operator const char *(),
				key.operator const char *());
	} else if ((!strcasecmp("multi",key)) | (!strcasecmp("multisession",key))) {
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		if (recorder != d->getType()) {
			errormsg("cannot set multisession flag: \"%s\" is not a recorder",
				dev.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		if ((!strcasecmp("no",value)) | (!strcasecmp("false",value))) {
			infomsg("setting multisession to false for \"%s\"",
				dev.operator const char *());
			r->setMultisession(false);
		} else if ((!strcasecmp("yes",value)) | (!strcasecmp("true",value))) {
			infomsg("setting multisession to true for \"%s\"",
				dev.operator const char *());
			r->setMultisession(true);
		} else
			errormsg("unexpected value \"%s\" for key \"%s\"",
				value.operator const char *(),
				key.operator const char *());
	} else if (!strcasecmp("speed",key)) {
		bool ok = false;
		int s = value.toInt(&ok);
		if (!ok) {
			errormsg("unexpected value \"%s\" for key \"%s\"",
				value.operator const char *(),
				key.operator const char *());
			return;
		}
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		CDROM *c = (CDROM *) d;
		switch (d->getType()) {
			case recorder:
				r->setWriteSpeed(s);
				break;
			case cdrom:
				c->setReadSpeed(s);
				break;
			default:
				errormsg("cannot set speed: \"%s\" is neither a cdrom nor a recorder.",
					dev.operator const char *());
		}
	} else if (!strcasecmp("driver",key)) {
		Device *d = getDevice(dev,bus,id,lun);
		if (!d) {
			errormsg("no such device \"%s\" (%s)",
				dev.operator const char *(),
				idStr.operator const char *());
			return;
		}
		if (recorder != d->getType()) {
			errormsg("Cannot set driver: \"%s\" is not a recorder.",
				dev.operator const char *());
			return;
		}
		Recorder *r = (Recorder *) d;
		QStringList::Iterator drv = drivers.begin();
		for (++drv; drv != drivers.end(); ++drv) {
			QString tmp = (*drv).mid(8);
			debugmsg("checking %s against %s", key.operator const char *(),
				tmp.operator const char *());
			if (!strcasecmp(tmp,value)) {
				r->setDriver(drivers.findIndex(*drv));
				return;
			}
		}
		errormsg("no driver \"%s\" in driver list",value.operator const char *());
	} else if (!strcasecmp("showlog",key)) {
		if (!boolvalue(value))
			mainWin->displayLog();	// hide log
	} else
		errormsg("unexpected key \"%s\" with value \"%s\" in config file: ingnored",
			key.operator const char *(),
			value.operator const char *());
}
