/*
 * 
 * 	fryit - file selections.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>

#include <qfile.h>
#include <qstrlist.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include <qtabbar.h>
#include <qtextstream.h>

#include "cdrom.h"
#include "dialogs.h"
#include "lowlevel.h"
#include "recorder.h"
#include "selections.h"
#include "track.h"
#include "widgets.h"

#include "globals.h"

ReadSelectionDialog::ReadSelectionDialog(CDROM *cdr) :
	QTabDialog(),
	dev(cdr),
	pr(0), first(true) {
	
	QString cap = cdr->name();
	cap += ": select tracks for reading";
	setCaption(cap);
	tracks.setAutoDelete(true);
	area = new QVBox(this);
	CHECK_PTR(area);
	area->setMargin(5);
	header = new QHeader(area);
	CHECK_PTR(header);
	
	setDefaultButton("Rescan CD");
	setCancelButton("Close");
	setOKButton("Read Tracks");
	trackBox = new QScrollView(area);
	CHECK_PTR(trackBox);
	trackBox->setVScrollBarMode(QScrollView::AlwaysOn);
	trackBox->setResizePolicy(QScrollView::Manual);
	showPage(trackBox);
	resize(600,size().height());
	connect(this,SIGNAL(defaultButtonPressed()),this,SLOT(rescanCD()));
	connect(this,SIGNAL(applyButtonPressed()),this,SLOT(checkTrackNames()));
	connect(this,SIGNAL(cancelButtonPressed()),SLOT(hide()));
	rdlg = 0;
	ddlg = 0;
	pr = 0;

	addTab(area,"Read Tracks");
	Track *tr = new Track(0,0,0,0,data1);
	TrackSelection *t = new TrackSelection(dev,trackBox,tr);
	QFont f("Helvetica",12);
	header->setFont(f);
	header->addLabel("read",t->getColumnSize(0) + 15);
	header->addLabel("Trackname",t->getColumnSize(1));
	header->addLabel("complete",t->getColumnSize(2) + 30);
	header->addLabel("begin at",t->getColumnSize(3) + 15);
	header->addLabel("end at",t->getColumnSize(4) + 15);
	header->setClickEnabled(false,-1);
	header->setClickEnabled(true,0);
	header->setResizeEnabled(false,-1);
	header->setMovingEnabled(false);
	connect(header,SIGNAL(sectionClicked(int)),SLOT(selectAll()));
	delete t;
	delete tr;
	
	resize(690,400);
}

void ReadSelectionDialog::showDialog() {
	first = true;
	rescanCD();
	show();
}

void ReadSelectionDialog::selectAll() {
	for (TrackSelection *e = tracks.first(); e; e = tracks.next())
		e->setSelected();
}

void ReadSelectionDialog::rescanCD() {
	bool m;
	try {
		m = dev->mediaChanged();
	} catch (Error e) {
		mainWin->postError(e);
		return;
	}
	if ((!m) & tracks.count())
		return;
	tracks.clear();
	QList <Track> cd;
	try {
		dev->getTOC(cd);
	}
	catch (Error e) {
		errormsg(e);
		if (isVisible()) {
			QMessageBox::critical(this,"error reading toc",
				"Could not read table of contents.\n"
				"Is there a CD in the drive?",
				"Ouch!");
			hide();
		}
	}
	int ypos = 0;
	for (Track *t = cd.first(); t; t = cd.next()) {
		TrackSelection *et = new TrackSelection(dev,trackBox,t);
		CHECK_PTR(et);
		et->resize(et->sizeHint());
		trackBox->resizeContents(et->width(), 
			ypos + et->sizeHint().height());
		trackBox->addChild(et,0,ypos);
		et->show();
		ypos += et->height();
		tracks.append(et);
	}
	trackBox->viewport()->repaint();
}

void ReadSelectionDialog::checkTrackNames() {
	QList <Track> trs;
	LowLevel::getDiscTracks(trs);
	QStrList names;
	debugmsg("existing tracks:");
	for (trs.first(); trs.current(); trs.next()) {
		names.append(trs.current()->getName());
		debugmsg("%s",trs.current()->getName());
	}
	for (TrackSelection *t = tracks.first(); t; t = tracks.next()) {
		if (!t->isSelected())
			continue;
		for (names.first(); names.current(); names.next())
			if (strcmp(names.current(),t->getName()) == 0) {
				QString str;
				str.sprintf("Cannot create track with name\n"
				"\"%s\".\n"
				"A track with this name already exists.",t->getName());
				QMessageBox::critical(0,"Error",str,"Retry");
				show();
				return;
			}
	}
	readTrack();
}

void ReadSelectionDialog::readTrack() {
	TrackSelection *t = tracks.current();
	if (first) {
		t = tracks.first();
		debugmsg("readTrack(): first call");
		first = false;
	} else {
		debugmsg("readTrack()");
		LowLevel::writeTrackInfo(t);
		if (rdlg) {
			dev->processFinished();
			rdlg->hide();
			delete rdlg;
			rdlg = 0;
			pr = 0;
		} else {
			ddlg->hide();
			delete ddlg;
			ddlg = 0;
		}
		mainWin->updateView();
		if ((t = tracks.next()) == 0)
			return;
	}
	while (!t->isSelected()) {
		t = tracks.next();
		if (t == 0) {
			hide();
			return;
		}
	}
	debugmsg("starting reading track \"%s\"",t->getName());
	if (data1 == t->getTrackType()) {
		ddlg = new ReadDataDialog(dev,t);
		CHECK_PTR(ddlg);
		ddlg->show();
		connect(ddlg,SIGNAL(finished()),SLOT(readTrack()));
		connect(ddlg,SIGNAL(cancelled()),SLOT(doCancel()));
		return;
	}
	
	rdlg = new ReadDialog(t->getName());
	CHECK_PTR(rdlg);
	rdlg->show();
	if (!((Track *) t)->getTrackID())
		t->setTrackID(LowLevel::newTrackID(audio == t->getTrackType()));
	pr = dev->readAudio(t);
	infomsg("writing Track %s (#%i) to file %s.cdr",t->getName(),
		t->getNumber(), t->filename());
	connect(pr,SIGNAL(out(const char *)),rdlg,SLOT(outMsg(const char *)));
	connect(pr,SIGNAL(err(const char *)),rdlg,SLOT(errMsg(const char *)));
	connect(pr,SIGNAL(quit(int)),SLOT(readTrack()));
	connect(rdlg,SIGNAL(cancelled()),SLOT(doCancel()));
}

void ReadSelectionDialog::doCancel() {
	if (rdlg) {
		dev->cancelAction();
		delete rdlg;
		rdlg = 0;
		pr = 0;
	} else {
		delete ddlg;
		ddlg = 0;
	}
}

void ReadSelectionDialog::resizeEvent(QResizeEvent *) {
	int w = trackBox->width() - trackBox->verticalScrollBar()->width();
	for (TrackSelection *t = tracks.first(); t; t = tracks.next())
		t->resize(w,t->height());
	TrackSelection *t;
	if (tracks.count())
		t = tracks.first();
	else {
		Track *tr = new Track(0,0,0,0,data1);
		t = new TrackSelection(dev,trackBox,tr);
		delete tr;
	}
	header->setCellSize(0,t->getColumnSize(0) + 15);
	header->setCellSize(1,t->getColumnSize(1));
	header->setCellSize(2,t->getColumnSize(2) + 30);
	header->setCellSize(3,t->getColumnSize(3) + 15);
	header->setCellSize(4,t->getColumnSize(4) + 15);
	if (!tracks.count())
		delete t;
	int h = header->sizeHint().height();
	header->setGeometry(0,0,area->width(),h);
	trackBox->setGeometry(0,h + 5,area->width(),area->height() - h - 5);
	trackBox->viewport()->repaint();
	header->repaint();
}

WriteSelectionDialog::WriteSelectionDialog(Recorder *rec) :
	QTabDialog(),
	dev(rec),
	burnDlg(rec->name()) {
	
	burnTracks.setAutoDelete(false);
	tabBar()->hide();
	
	QString cap;
	cap.sprintf("%s: select tracks for writing",dev->name());
	setCaption(cap);
	area = new QFrame(this);
	CHECK_PTR(area);
	area->setMargin(5);
	QFont f("Helvetica",14);
	mainLay = new QHBoxLayout(area);
	CHECK_PTR(mainLay);
	
	diskLay = new QVBoxLayout();
	CHECK_PTR(diskLay);
	writeLay = new QVBoxLayout();
	CHECK_PTR(writeLay);
	mainLay->addLayout(diskLay);
	mainLay->addLayout(writeLay);

	diskLabel = new QLabel("Tracks on harddisk",area);
	CHECK_PTR(diskLabel);
	diskLabel->resize(diskLabel->sizeHint());
	writeLabel = new QLabel("Tracks to write",area);
	CHECK_PTR(writeLabel);
	writeLabel->resize(writeLabel->sizeHint());
	diskLabel->setFont(f);
	diskLabel->setAlignment(AlignCenter);
	diskLabel->resize(diskLabel->sizeHint());
	writeLabel->setFont(f);
	writeLabel->setAlignment(AlignCenter);
	writeLabel->resize(writeLabel->sizeHint());
	diskLay->addWidget(diskLabel);
	diskLay->addSpacing(10);
	writeLay->addWidget(writeLabel);
	writeLay->addSpacing(10);
	
	diskList = new QListView(area);
	CHECK_PTR(diskList);
	diskList->addColumn("Trackname");
	diskList->addColumn("Size");
	diskList->setColumnAlignment(1,AlignCenter);
	diskList->header()->setFont(f);
	diskList->header()->setClickEnabled(false,1);
	diskList->header()->setMovingEnabled(false);
	diskList->setColumnWidth(0,100);
	diskList->setColumnWidth(1,50);
	diskList->setMultiSelection(true);
	diskLay->addWidget(diskList);
	diskLay->addSpacing(10);
	
	writeList = new QListView(area);
	CHECK_PTR(writeList);
	writeList->addColumn("Trackname");
	writeList->addColumn("Size");	
	writeList->setColumnAlignment(1,AlignCenter);
	writeList->header()->setClickEnabled(false);
	writeList->header()->setMovingEnabled(false);
	writeList->header()->setFont(f);
	writeList->setColumnWidth(0,100);
	writeList->setColumnWidth(1,50);
	writeList->setSorting(-1);		// disable sorting as we need
						// to know the desired order
	addTab(area,"write Tracks");
	showPage(area);
	setCancelButton("Cancel");
	setOKButton("Fry it!");
	writeLay->addWidget(writeList);
	writeLay->addSpacing(10);
	
	diskBtn = new QHBoxLayout();
	CHECK_PTR(diskBtn);
	diskBtn->setMargin(15);
	diskBtn->setSpacing(20);
	diskLay->addLayout(diskBtn);
	writeBtn = new QHBoxLayout();
	CHECK_PTR(writeBtn);
	writeBtn->setMargin(15);
	writeBtn->setSpacing(20);
	writeLay->addLayout(writeBtn);
	addBtn = new QPushButton("Add",area);
	CHECK_PTR(addBtn);
	remBtn = new QPushButton("Remove",area);
	CHECK_PTR(remBtn);
	delBtn = new QPushButton("Delete",area);
	CHECK_PTR(delBtn);
	upBtn = new QPushButton("Up",area);
	CHECK_PTR(upBtn);
	dwnBtn = new QPushButton("Down",area);
	CHECK_PTR(dwnBtn);
	addBtn->resize(addBtn->sizeHint());
	remBtn->resize(remBtn->sizeHint());
	upBtn->resize(dwnBtn->sizeHint());
	dwnBtn->resize(dwnBtn->sizeHint());
	delBtn->resize(delBtn->sizeHint());
	diskBtn->addWidget(addBtn);
	diskBtn->addWidget(delBtn);
	writeBtn->addWidget(upBtn);
	writeBtn->addWidget(dwnBtn);
	writeBtn->addWidget(remBtn);
	
	int h = diskLabel->sizeHint().height();
	h += diskList->sizeHint().height();
	h += addBtn->sizeHint().height();
	h += writeLabel->sizeHint().height();
	h += 20;
	area->setMinimumSize(diskList->sizeHint().width() * 2,h);
	
	connect(addBtn,SIGNAL(clicked()),SLOT(addTrack()));
	connect(remBtn,SIGNAL(clicked()),SLOT(remTrack()));
	connect(delBtn,SIGNAL(clicked()),SLOT(delTrack()));
	connect(upBtn,SIGNAL(clicked()),SLOT(upTrack()));
	connect(dwnBtn,SIGNAL(clicked()),SLOT(downTrack()));
	connect(diskList,SIGNAL(doubleClicked(QListViewItem *)),SLOT(addTrack()));
	connect(this,SIGNAL(applyButtonPressed()),SLOT(start()));
	connect(this,SIGNAL(cancelButtonPressed()),SLOT(hide()));
	connect(diskList->header(),SIGNAL(sectionClicked(int)),SLOT(selectAll()));
	connect(&burnDlg,SIGNAL(cancel()),SLOT(doCancel()));
	connect(&burnDlg,SIGNAL(cancel()),dev,SLOT(cancelWriting()));
	
	resize(550,400);
}

void WriteSelectionDialog::updateList() {
	diskTracks.clear();
	diskList->clear();
	writeList->clear();
	LowLevel::getDiscTracks(diskTracks);
	for (Track *tr = diskTracks.first(); tr; tr = diskTracks.next()) {
		debugmsg("inserting Track \"%s\" (file %s.cdr)",tr->getName(),
			tr->filename());
		TrackItem *it = new TrackItem(tr,diskList);
		CHECK_PTR(it);
	}
}

void WriteSelectionDialog::selectAll() {
	for (QListViewItem *it = diskList->firstChild(); it; it = it->nextSibling())
		diskList->setSelected(it,true);
}

void WriteSelectionDialog::addTrack() {
	for (TrackItem *it = (TrackItem *) diskList->firstChild(); it;
		 it = (TrackItem *) it->nextSibling()) {
		
		if (!it->isSelected())
			continue;
		diskList->setSelected(it,false);
		QListViewItem *i1 = writeList->firstChild();
		if (i1)
			while(i1->itemBelow())
				i1 = (TrackItem *) i1->itemBelow();
		debugmsg("adding track \"%s\" to writeList",it->tr->getName());
		TrackItem *i2 = new TrackItem(it->tr,writeList,i1);
		CHECK_PTR(i2);
		writeList->ensureItemVisible(i2);
	}
	calcLength();
}

void WriteSelectionDialog::remTrack() {
	QListViewItem *it;
	if ((it = writeList->currentItem()) == 0)
		return;
	debugmsg("removing track \"%s\" from writing list",it->text(1).operator const char *());
	delete it;
	calcLength();
}

void WriteSelectionDialog::delTrack() {
	QListViewItem *it;
	if ((it = diskList->currentItem()) == 0)
		return;
	Track *tr;
	for (tr = diskTracks.first(); tr; tr = diskTracks.next())
		if (strcmp(tr->getName(),it->text(0)) == 0)
			break;
	diskTracks.find(tr);
	LowLevel::deleteTrack(tr);
	for (QListViewItem *i2 = writeList->firstChild(); i2; i2 = i2->nextSibling())
		if (strcmp(i2->text(0),it->text(0)) == 0)
			delete i2;
	delete it;
	diskTracks.remove(tr);
}

void WriteSelectionDialog::upTrack() {
	TrackItem *it;
	if ((it = (TrackItem *) writeList->currentItem()) == 0)
		return;
	QListViewItem *after = it->itemAbove();
	if (!after)	// is already top item
		return;
	after = after->itemAbove();
	TrackItem *n = new TrackItem(it->tr,writeList,after);
	CHECK_PTR(n);
	delete it;
}

void WriteSelectionDialog::downTrack() {
	TrackItem *it;
	if ((it = (TrackItem *) writeList->currentItem()) == 0)
		return;
	QListViewItem *after = it->itemBelow();
	if (!after)	// is already last item
		return;
	TrackItem *n = new TrackItem(it->tr,writeList,after);
	CHECK_PTR(n);
	writeList->setSelected(n,true);
	writeList->ensureItemVisible(n);
	delete it;
}

void WriteSelectionDialog::start() {
	if ((min > 74) | ((min == 74) & ((sec > 0) | (hun > 0))))
		if (1 == QMessageBox::warning(this,"Overlength",
			"You have selected more than 74 minutes!\n"
			"Do you want to continue?","Yes","No"))
			return;
	debugmsg("WriteSelectionDialog::start(): tracks selected:");
	TrackItem *it;
	total = 0;
	burnTracks.clear();
	for (it = (TrackItem *) writeList->firstChild(); it; it = (TrackItem *) it->nextSibling()) {
		burnTracks.append(it->tr);
		infomsg("\"%s\" selected for writing (file %s.cdr)",
			it->tr->getName(),it->tr->filename());
		total += it->tr->getLength();
	}
	total /= 1024*1024;
	if (0 == burnTracks.count())
		return;
	QMessageBox m("Insert CD-R",
		"Please insert a CD-R in your cd writer.",
		QMessageBox::Information,1,0,0);
	QFont f("Helvetica",12);
	m.setFont(f);
	m.exec();

	written = 0;
	pr = dev->writeTracks(burnTracks);
	if (!pr)
		return;
	burnDlg.setInfo("starting recorder...");
	burnDlg.reset();
	burnDlg.setFifo(100);
	burnDlg.show();
	connect(pr,SIGNAL(out(const char *)),SLOT(outMsg(const char *)));
	connect(pr,SIGNAL(err(const char *)),SLOT(errMsg(const char *)));
	connect(pr,SIGNAL(quit(int)),SLOT(finish()));
	connect(pr,SIGNAL(quit(int)),dev,SLOT(writingFinished()));
	hide();
}

void WriteSelectionDialog::calcLength() {
	debugmsg("WriteSelectionDialog::calcLength()");
	min = sec = hun = 0;
	for (TrackItem *t = (TrackItem *) writeList->firstChild(); t;
		t = (TrackItem *) t->nextSibling()) {
		int m = 0, s = 0, h = 0;
		t->tr->getTime(m,s,h);
		min += m;
		sec += s;
		hun += h;
		debugmsg("got getTime(%i,%i,%i)",m,s,h);
	}
	sec += hun/100;
	hun %= 100;
	min += sec/60;
	sec %= 60;
	QString str;
	str.sprintf("Tracks to write (%i:%02i.%02i)",min,sec,hun);
	writeLabel->setText(str);
}

void WriteSelectionDialog::outMsg(const char *msg) {
	int tr,f;
	QString m = msg;
	QTextStream ts(m,IO_ReadOnly);
	while (!ts.atEnd()) {
		QString line = ts.readLine().simplifyWhiteSpace();
		debugmsg("analyzing line \"%s\"",line.operator const char *());
		if (line.isNull())
			continue;
		if (2 == sscanf(line,"Track %i: %*i of %*i MB written (fifo %i%%).",&tr,&f)) {
			written++;
			burnDlg.setProgress((int) rint((float)written / (float)total * 100));
			burnDlg.setFifo(f);
			QString str;
			str.sprintf("Burning Track %i...",tr);
			burnDlg.setInfo(str);
		} else if (!strncmp(line,"Waiting for reader process to fill input-buffer",47))
			burnDlg.setInfo("filling input buffer...");
		else if (!strncmp(line,"Fixating...",11))
			burnDlg.setInfo("Fixating...");
	}
}

void WriteSelectionDialog::errMsg(const char *m) {
	QString str = m;
	QTextStream ts(str,IO_ReadOnly);
	while (!ts.atEnd()) {
		QString msg = ts.readLine();
		msg = msg.mid(str.find(' ') + 1);
		if (-1 != msg.find("fifo"))
			infomsg("%s",msg.operator const char *());
		else
			warningmsg("cdrecord wrote to stderr:\n%s",msg.operator const char *());
	}
}

void WriteSelectionDialog::finish() {
	debugmsg("WriteSelectionDialog::finish()");
	burnDlg.hide();
}

void WriteSelectionDialog::doCancel() {
	debugmsg("WriteSelectioinDialog::doCancel()");
	pr = 0;
	burnDlg.hide();
}

void WriteSelectionDialog::show() {
	updateList();
	QDialog::show();
}

CopySelectionDialog::CopySelectionDialog() :
	QDialog(),
	pr(0) {
	
	setCaption("Copy CD");
	
	area = new QGridLayout(this,4,2,5,3);
	CHECK_PTR(area);
	readerLbl = new QLabel("Reader: ",this);
	CHECK_PTR(readerLbl);
	writerLbl = new QLabel("Writer: ",this);
	CHECK_PTR(writerLbl);
	reader = new QComboBox(this);
	CHECK_PTR(reader);
	writer = new QComboBox(this);
	CHECK_PTR(writer);
	area->addWidget(readerLbl,0,0);
	area->addWidget(writerLbl,1,0);
	area->addWidget(reader,0,1);
	area->addWidget(writer,1,1);
	area->setColStretch(0,1);
	area->setColStretch(1,3);
	
	buttonArea = new QHBoxLayout(12);
	CHECK_PTR(buttonArea);
	ok = new QPushButton("OK",this);
	CHECK_PTR(ok);
	cancel = new QPushButton("Cancel",this);
	CHECK_PTR(cancel);
	buttonArea->addSpacing(10);
	buttonArea->addWidget(ok);
	buttonArea->addWidget(cancel);
	buttonArea->addSpacing(11);
	
	area->addMultiCell(buttonArea,3,3,0,1);
	
	connect(ok,SIGNAL(clicked()),SLOT(start()));
	connect(cancel,SIGNAL(clicked()),SLOT(hide()));
	
	resize(250,100);
}

void CopySelectionDialog::start() {
	hide();
	if (0 != QMessageBox::warning(this,"Warning",
		"This operation might easily\n"
		"fail if your reader isnt fast enough.\n"
		"Continue only in simulation mode or if\n"
		"you like to take some risk...\n","Continue","Cancel"))
		return;
	rec = RecorderList.at(writer->currentItem());
	cdr = CdromList.at(reader->currentItem());
	if (CDS_MIXED == cdr->getCDtype()) {
		QMessageBox::critical(this,"Mixed Mode CD",
			"Cannot read data from a mixed mode CD!\n"
			"Please wait for a new version of this programm.",
			"Sorry");
		return;
	}
	try {
		size = cdr->getDataSize();
	}
	catch (Error e) {
		mainWin->postError(e);
		return;
	}
	pr = rec->writeThroughStdin(size);
	cdr->readToProcess(size,pr);
	connect(pr,SIGNAL(out(const char *)),SLOT(outMsg(const char *)));
	connect(pr,SIGNAL(err(const char *)),SLOT(errMsg(const char *)));
	connect(pr,SIGNAL(quit(int)),SLOT(finish()));
	burnDlg = new BurnDialog("Copy CD");
	CHECK_PTR(burnDlg);
	burnDlg->show();
	burnDlg->setInfo("starting recorder...");
	connect(burnDlg,SIGNAL(cancel()),SLOT(doCancel()));
}

void CopySelectionDialog::show() {
	reader->clear();
	writer->clear();
	for (Recorder *d = RecorderList.first(); d; d = RecorderList.next())
		if (idle == d->getStatus())
			writer->insertItem(d->name());
	for (CDROM *d = CdromList.first(); d; d = CdromList.next())
		if (idle == d->getStatus())
			reader->insertItem(d->name());
	QDialog::show();
}

void CopySelectionDialog::outMsg(const char *msg) {
	int pr,s,f;
	QString str = msg;
	str = str.simplifyWhiteSpace();
	debugmsg("outMsg(\"%s\")", str.operator const char *());
	QRegExp r("Track\\s*[0-9]+:\\s*[0-9]+ of\\s*[0-9]+ MB written");
	int at = str.find(r);
	if (-1 != at) {
		if (3 == sscanf(str.mid(at),"Track %*i: %i of %i MB written (fifo %i%%).",&pr,&s,&f)) {
			int prog = (int) rint((float) pr / (float) s * 100);
			burnDlg->setProgress(prog);
			burnDlg->setFifo(f);
		} else if (2 == sscanf(str.mid(at),"Track %i: %*i of %i MB written.",&pr,&s)) {
			burnDlg->setInfo("Copying CD...");
			burnDlg->setFifo(100);
		}
	} else if (-1 != str.find("Fixating..."))
		burnDlg->setInfo("Fixating...");
}

void CopySelectionDialog::errMsg(const char *msg) {
	warningmsg("cdrecord wrote to stderr: %s",msg);
}

void CopySelectionDialog::finish() {
	burnDlg->hide();
	delete burnDlg;
	delete pr;
	delete this;
}

void CopySelectionDialog::doCancel() {
	rec->cancelWriting();
	cdr->cancelAction();
}

#include "selections.moc"
