/*
 * 
 * 	fryit - file threads.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 	
 */

#include "config.h"

#include "device.h"
#include "cdrom.h"
#include "globals.h"
#include "threads.h"
#include "track.h"

#include <cerrno>
#include <cmath>
#include <cstdio>
#include <fcntl.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>

void threadTerminated(void *arg) {
	debugmsg("threadTerminated(...) called");
	Device *d = (Device *) arg;
	d->setStatus(idle);
	d->thrFinished = true;
}

void *readData(void **arg) {
	CDROM *cdr = (CDROM *) arg[0];
	int track = (int) arg[1];
	const char *filename = (const char *) arg[2];
	int startOff = (int) arg[3];	// these values are relative
	int endOff = (int) arg[4];
	debugmsg("readData(CDROM, %i, \"%s\", %i, %i)",track,filename,startOff,endOff);

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,0);
	pthread_cleanup_push(threadTerminated,cdr);
	try {
		File dev(cdr->bDev, O_RDONLY | O_NONBLOCK, true);
		cdrom_tochdr hdr;
		if (ioctl(dev,CDROMREADTOCHDR,&hdr) == -1) {
			QString err;
			err.sprintf("could not get toc: %s",strerror(errno));
			mainWin->postError(err);
			pthread_exit((void *) -1);
		}
		if ((track < hdr.cdth_trk0) || (track > hdr.cdth_trk1)) {
			QString err;
			err.sprintf("track index %i out of range [%i,%i]",
				track,
				(int) hdr.cdth_trk0,
				(int) hdr.cdth_trk1);
			mainWin->postError(err);
			pthread_exit((void *) -1);
		}
		File out(filename, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
		cdrom_tocentry entr;
		entr.cdte_track = track;
		entr.cdte_format = CDROM_LBA;
		if (ioctl(dev,CDROMREADTOCENTRY,&entr) == -1) {
			QString err;
			err.sprintf("could not get track info: %s",strerror(errno));
			mainWin->postError(err);
			pthread_exit((void *) -1);
		}
		int start = entr.cdte_addr.lba;		// absolute position
		entr.cdte_track = (track < hdr.cdth_trk1) ? track + 1 : CDROM_LEADOUT;
		if (ioctl(dev,CDROMREADTOCENTRY,&entr) == -1) {
			QString err;
			err.sprintf("could not get track info: %s",strerror(errno));
			mainWin->postError(err);
			pthread_exit((void *) -1);
		}
		int end = entr.cdte_addr.lba;
		debugmsg("Track %i at %i - ending at %i",track,start,end);
		if ((end - start < endOff/2) | (end - start < startOff/2)) {
			QString err;
			err.sprintf("Selected Offsets are out of range!\n"
				"This is a bug! Please report this.");
			mainWin->postError(err);
			pthread_exit((void *) -1);
		}
		debugmsg("seeking to position %i", startOff * 1024);
		
		long startpos = 0;
		int laststart;
		entr.cdte_addr.lba = 0;
		for (int c = 1; c <= track; c++) {
			laststart = entr.cdte_addr.lba;
			entr.cdte_track = c;
			if (-1 == ioctl(dev,CDROMREADTOCENTRY,&entr)) {
				QString err;
				err.sprintf("could not get track info: %s",strerror(errno));
				mainWin->postError(err);
				pthread_exit((void *) -1);
			}
			int tmp = entr.cdte_addr.lba - laststart;
			debugmsg("track %i: %i blocks",c,tmp);
			tmp *= (entr.cdte_datamode == audio) ? 2646 : 2048;
			debugmsg("adding %i to startpos",tmp);
			startpos += tmp;
		}
		startpos += startOff * 1024;
		debugmsg("startpos = %i",startpos);
		lseek(dev, startpos, SEEK_SET);
		
		const int blocksize = 512 * 1024;	// performance tuning here
		char buf[blocksize];
		int i = blocksize;
		int num = (endOff - startOff) << 10;
		int total = num;
		debugmsg("selected length: %i kByte",num >> 10);
		while (num > 0) {
			cdr->thrProgress = (int) rint((float) (total - num) / (float) total * (float) 100);
			if (blocksize > num)
				i = num;
			num -= blocksize;
			if (read(dev,buf,i) == -1) {
				QString err;
				err.sprintf("error reading track: %s\n",strerror(errno));
				mainWin->postError(err);
				pthread_exit((void *) -1);
			}
			if (write(out,buf,i) == -1) {
				QString err;
				err.sprintf("error writing data to disk: %s\n",strerror(errno));
				mainWin->postError(err);
				pthread_exit((void *) -1);
			}
		}
		debugmsg("finished reading track");
	}
	catch (Error e) {
		mainWin->postError(e);
	}
	pthread_cleanup_pop(1);
	pthread_exit(0);
}

void *readDataToProcess(void **arg) {
	CDROM *cdr = (CDROM *) arg[0];
	int size = (int) arg[1];
	Process *pr = (Process *) arg[2];

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,0);
	pthread_cleanup_push(threadTerminated,(void *) cdr);

	try {
		File dev(cdr->bDev, O_RDONLY | O_NONBLOCK, true);
		char buf[512 * 1024];
		int num,total = 0;
		lseek(dev,0,SEEK_SET);
		while ((num = read(dev,buf,512 * 1024)) > 0) {
			total += num;
			if (num == -1) {
				QString err = "Could not read data from CD: ";
				err += strerror(errno);
				mainWin->postError(err);
				pthread_exit((void *) -1);
			}
			pr->input(buf,num);
			if (total >= size)
				break;
		}		
	}
	catch (Error e) {
		mainWin->postError(e);
		pr->endinput();
		pthread_exit((void *) -1);
	}	
	pr->endinput();
	pthread_cleanup_pop(1);
	pthread_exit(0);
}

