/*
 * 
 * 	fryit - file widgets.cc
 * 	Copyright Thomas Maier-Komor, 1999
 * 	
 */

#include "config.h"

#include <cassert>

#include <qmessagebox.h>

#include "cdrom.h"
#include "widgets.h"
#include "track.h"

#include "globals.h"

TrackItem::TrackItem(const Track *t, QListView *p = 0, QListViewItem *after = 0) :
	QListViewItem(p,after),
	tr(t) {
	
	debugmsg("TrackItem created for track %s",tr->getName());
}

QString TrackItem::text(int c) const {
	switch (c) {
		case 0:
			return tr->getName();
		case 1:
			switch (tr->getTrackType()) {
				case audio:
					int min,sec,hun;
					tr->getTime(min,sec,hun);
					col1.sprintf("%i:%02i:%02i",min,sec,hun);
					return col1;
				case data1:
					col1.sprintf("%i kByte",tr->getLength() / 1024);
					return col1;
				default:
					assert(0);
			}
		case 2:
			switch (tr->getTrackType()) {
				case audio:
					return "audio";
				case data1:
					return "data";
				default:
					assert(0);
			}
		default:
			assert(0);
	}
	return 0;
}

TrackSelection::TrackSelection(CDROM *cdr, QWidget *parent, const Track *tr) :
	QHBox(parent), Track(*tr),
	cdrom(cdr) {
	
	setSpacing(5);
	selectBox = new QCheckBox(this);
	CHECK_PTR(selectBox);
	nameEdit = new QLineEdit(this);
	CHECK_PTR(nameEdit);
	nameEdit->setMaxLength(100);
	nameEdit->setText(getName());
	completeBox = new QCheckBox(this);
	CHECK_PTR(completeBox);
	
	if (audio == getTrackType()) {
		int m = 0, s = 0, h = 0;
		getTime(m,s,h);
		sT = new TimeBox(0,0,0,m,s,h,this);
		CHECK_PTR(sT);
		sT->setCurrent(0,0,0);
		space = new QLabel(" ",this);
		CHECK_PTR(space);
		eT = new TimeBox(0,0,0,m,s,h,this);
		CHECK_PTR(eT);
		eT->setCurrent(m,s,h);
		sO = eO = 0;
		playBtn = new QPushButton("play",this);
		CHECK_PTR(playBtn);
		connect(playBtn,SIGNAL(clicked()),SLOT(play()));
		connect(sT,SIGNAL(posChanged()),SLOT(checkTimes()));
		connect(eT,SIGNAL(posChanged()),SLOT(checkTimes()));
	} else {	// data1 == getTrackType() assumed
		sO = new QSpinBox(0, getLength() / 1024, 1, this);
		CHECK_PTR(sO);
		sO->setValue(0);
		space = new QLabel(" ",this);
		CHECK_PTR(space);
		eO = new QSpinBox(0, getLength() / 1024, 1, this);
		CHECK_PTR(eO);
		eO->setValue(getLength() / 1024);
		sT = eT = 0;
		playBtn = 0;
		connect(sO,SIGNAL(valueChanged(int)),SLOT(checkOffsets()));
		connect(eO,SIGNAL(valueChanged(int)),SLOT(checkOffsets()));
	}
	
	connect(completeBox,SIGNAL(toggled(bool)),SLOT(setComplete(bool)));
	setMinimumSize(630,
		audio == getTrackType() ? sT->sizeHint().height(): sO->sizeHint().height());
	setComplete(true);
	
	connect(nameEdit,SIGNAL(textChanged(const QString &)),SLOT(setName(const QString &)));
}

QSize TrackSelection::sizeHint() const {
	QSize s;
	s.setWidth(620);
	s.setHeight(sT ? sT->sizeHint().height() : sO->sizeHint().height() + 2);
	return s;
}

void TrackSelection::setName(const QString &name) {
	nameEdit->setText(name);
	Track::setName(name);
}

void TrackSelection::setComplete(bool c) {
	completeBox->setChecked(c);
	if (c) {
		if (sT) {
			int min,sec,hun;
			getTime(min,sec,hun);
			sT->setCurrent(0,0,0);
			eT->setCurrent(min,sec,hun);
			sT->lock();
			eT->lock();
		} else {
			sO->setRange(0,0);
			eO->setRange(getLength() / 1024,getLength() / 1024);
			sO->setValue(0);
			eO->setValue(getLength() / 1024);
		}
	} else {
		if (sT) {
			sT->lock(false);
			eT->lock(false);
		} else {
			sO->setRange(0,getLength() / 1024);
			eO->setRange(0,getLength() / 1024);
		}
	}
}

int TrackSelection::getColumnSize(int c) const {
	switch (c) {
		case 0:
			return selectBox->width();
		case 1:
			return nameEdit->width();
		case 2:
			return completeBox->width();
		case 3:
			return sT ? sT->width() : sO->width();
		case 4: 
			return eT ? eT->width() : eO->width();
		default:
			assert(0);
	}
	return 0;
}

void TrackSelection::play() {
	debugmsg("play track %i",getNumber());
	try {
		cdrom->play(getNumber());
	}
	catch (Error e) {
		QString cap = "Error";
		QMessageBox::critical(0,cap,e);
	}
}

void TrackSelection::checkOffsets() {
	if (sO->value() > eO->value())
		eO->setValue(sO->value());
}

void TrackSelection::checkTimes() {
	if (sT->pos() > eT->pos())
		eT->setPos(sT->pos());
}

void TrackSelection::getSelectedTime(int &M, int &S, int &H) const {
	assert(sT);
	int m,s,h;
	eT->getCurrent(M,S,H);
	sT->getCurrent(m,s,h);
	M -= m;
	S -= s;
	H -= h;
	if (H < 0) {
		H += 100;
		S--;
	}
	if (S < 0) {
		S += 60;
		M--;
	}
}

TimeBox::TimeBox(int m,int s,int h,int M,int S,int H,QWidget *p)
	: QHBox(p) {
	
	minMin = m;
	minSec = s;
	minHun = h;
	maxMin = M;
	maxSec = S;
	maxHun = H;
	
	min = new QSpinBox(0,M,1,this);
	CHECK_PTR(min);
	min->setWrapping(true);
	min->setPrefix("0");
	l0 = new QLabel(":",this);
	CHECK_PTR(l0);
	l0->setAlignment(AlignCenter);
	sec = new QSpinBox(0, M ? 59 : S, 1, this);
	CHECK_PTR(sec);
	sec->setWrapping(true);
	sec->setPrefix("0");
	l1 = new QLabel(".",this);
	CHECK_PTR(l1);
	l1->setAlignment(AlignCenter);
	hun = new QSpinBox(0, (M || S) ? 99 : H,1,this);
	CHECK_PTR(hun);
	hun->setWrapping(true);
	hun->setPrefix("0");
	ls = lh = 0;
	
	connect(min,SIGNAL(valueChanged(int)),this,SLOT(checkMin(int)));
	connect(sec,SIGNAL(valueChanged(int)),this,SLOT(checkSec(int)));
	connect(hun,SIGNAL(valueChanged(int)),this,SLOT(checkHun(int)));
	locked = false;
}

QSize TimeBox::sizeHint() const {
	QSize s;
	s.setWidth(min->sizeHint().width()*3 + 15);
	s.setHeight(min->sizeHint().height());
	return s;
}

void TimeBox::lock(bool l = true) {
	locked = l;
	if (l) {
		min->setRange(min->value(),min->value());
		sec->setRange(sec->value(),sec->value());
		hun->setRange(hun->value(),hun->value());
	} else {
		min->setRange(0,99);
		sec->setRange(0,59);
		hun->setRange(0,99);
	}
}

void TimeBox::setMinimum(int m,int s,int h) {
	minMin = m;
	minSec = s;
	minHun = h;
}

void TimeBox::setMaximum(int m,int s,int h) {
	maxMin = m;
	maxSec = s;
	maxHun = h;
}

void TimeBox::setCurrent(int m,int s,int h) {
	if ((m > maxMin) | (m < minMin))
		return;
	if ((s > maxSec) | (s < minSec))
		return;
	if ((h > maxHun) | (h < minHun))
		return;
	min->setValue(m);
	sec->setValue(s);
	hun->setValue(h);
	lm = m;
	ls = s;
	lh = h;
}

void TimeBox::getCurrent(int &m,int &s,int &h) const {
	m = min->value();
	s = sec->value();
	h = hun->value();
}

void TimeBox::checkMin(int m) {
	if (locked) {
		min->setValue(lm);
		return;
	}
	if (m > maxMin)
		min->setValue(maxMin);
	else if (m < minMin)
		min->setValue(minMin);
	else if (m == maxMin)
		checkSec(ls);
	lm = m;
	emit timeChanged(min->value(),sec->value(),hun->value());
	emit posChanged();
}

void TimeBox::checkSec(int s) {
	if (locked) {
		sec->setValue(ls);
		return;
	}
	if (min->value() == maxMin)
		if (s > maxSec)
			sec->setValue(0);
	if (min->value() == minMin)
		if (s < minSec)
			sec->setValue(minSec);
	if ((s == 0) && (ls == 59))
		min->setValue(min->value() + 1);
	else if ((s == 59) && (ls == 0))
		min->setValue(min->value() - 1);
	if (s < 10)
		sec->setPrefix("0");
	else
		sec->setPrefix("");
	if (s == maxSec)
		checkHun(lh);
	ls = s;
	emit timeChanged(min->value(),sec->value(),hun->value());
	emit posChanged();
}

void TimeBox::checkHun(int h) {
	if (locked) {
		hun->setValue(lh);
		return;
	}
	if (min->value() == maxMin)
		if (sec->value() == maxSec)
			if (h > maxHun)
				hun->setValue(0);
	if (min->value() == minMin)
		if (sec->value() == minSec)
			if (h < minHun)
				hun->setValue(minHun);
	if ((h == 0) && (lh == 99))
		min->setValue(min->value() + 1);
	if ((h == 99) && (lh == 0))
		min->setValue(min->value() - 1);
	if (h < 10)
		hun->setPrefix("0");
	else
		hun->setPrefix("");
	lh = h;
	emit timeChanged(min->value(),sec->value(),hun->value());
	emit posChanged();
}

void TimeBox::setPos(int h) {
	int m = h/6000;
	h -= (h/6000) * 6000;
	int s = h/100;
	h -= (h/100) * 100;
	setCurrent(m,s,h);
}

#include "widgets.moc"
